import logging

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import validate_slug
from django.utils.translation import gettext_lazy as _

from core.base_forms import TranslatedFieldsForm
from core.models import (
    Application,
    Assembly,
    AssemblyMember,
)

logger = logging.getLogger(__name__)


class AssemblyCreateForm(forms.ModelForm):
    class Meta:
        model = Assembly
        fields = ['slug', 'name']

    def __init__(self, *args, **kwargs):
        self._conference = kwargs.pop('conference')
        super().__init__(*args, **kwargs)

        if self._conference.has_disclaimer('assembly'):
            self.fields['disclaimer'] = forms.BooleanField(
                initial=False,
                required=False,
                label=self._conference.get_disclaimer('assembly'),
            )

    def clean(self, *args, **kwargs):
        super().clean(*args, **kwargs)

        if self._conference.has_disclaimer('assembly') and not self.cleaned_data['disclaimer']:
            self.add_error('disclaimer', _('Assembly__disclaimer-needed'))

        if (slug := self.cleaned_data.get('slug')) and Assembly.objects.filter(conference=self._conference, slug=slug).exists():
            self.add_error('slug', _('Assembly__slug__alreadyexists'))

        # ensure Assembly's conference is set before saving
        self.instance.conference = self._conference


class AssemblyEditForm(TranslatedFieldsForm):
    class Meta:
        model = Assembly
        fields = [
            'slug',
            'name',
            'is_physical',
            'is_virtual',
            'is_remote',
            'banner_image',
            'description',
            'assembly_link',
            'assembly_location',
            'registration_details',
        ]
        widgets = {
            'assembly_location': forms.Textarea(attrs={'rows': 4}),
        }

    def __init__(self, *args, staff_access: bool, staff_mode: bool, assembly_staff_access: bool, **kwargs):
        super().__init__(*args, **kwargs)

        # configure fields' widget customizations
        self.fields['assembly_link'].widget.attrs['placeholder'] = 'https://'

        # configure fields
        assembly = self.instance
        conference = self.instance.conference
        if assembly.base_data_readonly and not staff_access:
            self.fields['name'].disabled = True
            self.fields['slug'].disabled = True
            self.fields['is_physical'].disabled = True
            self.fields['is_remote'].disabled = True
            self.fields['is_virtual'].disabled = True

        if not conference.support_assembly_physical:
            self.fields['is_physical'].disabled = True

        if not conference.support_assembly_virtual:
            self.fields['is_virtual'].disabled = True

        if not conference.support_assembly_remote:
            self.fields['is_remote'].disabled = True
            self.fields['assembly_location'].disabled = True

        # disable all three fields if only one of physical/virtual/remote is available, no matter if staff or not
        ssat = conference.get_single_supported_assembly_type()
        if ssat is not None:
            self.fields['is_physical'].disabled = True
            self.fields['is_virtual'].disabled = True
            self.fields['is_remote'].disabled = True

    tags = forms.CharField(required=False)

    def clean_tags(self):
        # try to detect people delimiting tags with space instead of comma

        tags = self.cleaned_data['tags']
        # allow empty tags
        if tags.strip() == '':
            return ''

        split_tags = tags.split(',')
        if tags is not None and len(split_tags) == 1 and len(tags.split(' ')) > 2:
            raise ValidationError(_('Assembly__tags__splitwithcomma'))

        for tag in split_tags:
            stripped_tag = tag.strip()
            validate_slug(stripped_tag)

        return tags

    def clean(self):
        # call original .clean() which e.g. removes 'slug' from cleaned_data if that isn't a slug
        super().clean()

        # slug must not already exist in the conference
        slug = self.cleaned_data.get('slug')
        if slug is not None and Assembly.objects.filter(conference=self.instance.conference, slug=slug).exclude(pk=self.instance.pk).exists():
            self.add_error('slug', _('Assembly__slug__alreadyexists'))


class AssemblyApplicationAddForm(forms.Form):
    name = forms.CharField(label='Name')
    client_type = forms.ChoiceField(choices=Application.CLIENT_TYPES, initial=Application.CLIENT_CONFIDENTIAL)
    grant_type = forms.ChoiceField(choices=Application.GRANT_TYPES, initial=Application.GRANT_AUTHORIZATION_CODE)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['name'].widget.attrs['placeholder'] = 'Name'
        self.fields['client_type'].widget.attrs['placeholder'] = 'Client Type'
        self.fields['grant_type'].widget.attrs['placeholder'] = 'Authorization Grant Type'


class AssemblyMemberAddForm(forms.Form):
    name = forms.CharField()


class AssemblyMemberEditForm(forms.ModelForm):
    class Meta:
        model = AssemblyMember
        fields = ['is_representative', 'can_manage_assembly', 'is_technical_contact', 'show_public']

    def clean(self):
        # call original .clean() which does some checks already
        super().clean()

        a = self.instance.assembly
        m = self.instance.member
        if not self.cleaned_data['can_manage_assembly'] and not a.members.exclude(member=m).filter(can_manage_assembly=True):
            self.add_error('can_manage_assembly', _('AssemblyMember__needamanager'))
