import logging
from typing import Any

from django import forms
from django.utils.translation import gettext_lazy as _

from core.base_forms import TranslatedFieldsForm
from core.models import (
    Assembly,
    Event,
    PlatformUser,
    Room,
)
from core.models.tags import clean_tags

logger = logging.getLogger(__name__)


class EventForm(TranslatedFieldsForm):
    class Meta:
        model = Event
        fields = [
            'room',
            'location',
            'name',
            'language',
            'banner_image',
            'is_public',
            'abstract',
            'description',
            'schedule_start',
            'schedule_duration',
            'tags_list',
        ]
        widgets = {
            'abstract': forms.Textarea(attrs={'rows': 4}),
            'language': forms.Select(choices=[(x, x) for x in ['', 'de', 'en', 'de, en', 'en, de']]),  # TODO: proper solution :D
            'is_public': forms.HiddenInput(),
            'schedule_start': forms.DateTimeInput(attrs={'type': 'datetime-local'}, format='%Y-%m-%dT%H:%M:%S%z'),
        }

    tags_list = forms.CharField(required=False, label=_('Tags'), help_text=_('Project__tags__help'))

    def __init__(
        self,
        *args,
        conference,
        assembly: Assembly | None = None,
        owner: PlatformUser | None = None,
        create: bool = False,
        publish: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.conference = conference
        self.create = create
        self.sos = not assembly
        self.event_type_name = _('SoS') if self.sos else _('event')
        self.owner = owner
        self.publish = publish
        if self.sos:
            del self.fields['banner_image']
            del self.fields['abstract']
            self.fields['is_public'].initial = True
            self.assembly = self.conference.self_organized_sessions_assembly
            self.kind = Event.Kind.SELF_ORGANIZED
        else:
            self.assembly = assembly
            self.kind = Event.Kind.ASSEMBLY
        self.fields['schedule_start'].widget.attrs['placeholder'] = _('Event__schedule_start__placeholder')

        self._rooms = Room.objects.assignable_in_timeframe(
            conference=conference,
            start=self.instance.schedule_start,
            end=self.instance.schedule_end,
            assembly=self.assembly,
            is_sos=self.sos,
        )
        self.fields['room'].queryset = self._rooms
        self.fields['room'].label = _('Event__room')
        if self.instance.pk:
            self.initial['tags_list'] = ', '.join(self.instance.sorted_tags)

    def clean_tags_list(self):
        return clean_tags(self.cleaned_data['tags_list'])

    def clean(self):
        if self.cleaned_data.get('schedule_duration', None) is None:
            self.instance.schedule_end = None
        self.cleaned_data['is_public'] = not self.instance.is_public if self.publish else self.instance.is_public
        self.instance.conference = self.conference
        self.instance.assembly = self.assembly
        self.instance.kind = self.kind
        if self.owner and not self.instance.owner:
            self.instance.owner = self.owner

        return super().clean()

    def save(self, commit: bool = True) -> Any:
        event: Event = super().save(False)
        event.update_tags(self.cleaned_data['tags_list'], auto_create=True)
        if commit:
            event.save()
        return event
