import logging

from rest_framework.authtoken.models import Token

from django.contrib import messages
from django.shortcuts import redirect
from django.utils.html import format_html
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView, View
from django.views.generic.edit import FormView, UpdateView

from core.models.sso import Application
from core.models.voucher import Voucher

from backoffice.forms import (
    AssemblyApplicationAddForm,
)
from backoffice.views.mixins import AssemblyMixin

logger = logging.getLogger(__name__)


class AuthView(AssemblyMixin, FormView):
    form_class = AssemblyApplicationAddForm
    template_name = 'backoffice/assembly_auth.html'
    assembly_url_param = 'assembly'
    assembly_management = True

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        qs = Application.objects.filter(assembly=self.assembly)
        ctx['applications'] = qs.values('id', 'name', 'client_id', 'redirect_uris', 'client_type', 'authorization_grant_type')
        return ctx

    def form_valid(self, form):
        data = form.cleaned_data
        try:
            app = Application(
                assembly=self.assembly,
                user=self.assembly.create_technical_user_if_necessary(),
                name=data['name'],
                client_type=data['client_type'],
                authorization_grant_type=data['grant_type'],
            )
            app_client_secret = app.client_secret
            app.save()
            messages.success(
                self.request,
                format_html(
                    '{msg}:<br><strong><code>{secret}</code></strong>',
                    msg=_('Application__newclientsecret'),
                    secret=app_client_secret,
                ),
            )
            logger.info(
                'New OAuth2 app "%(app_name)s" created for assembly %(assembly)s by %(user)s',
                {'app_name': app.name, 'assembly': self.assembly, 'user': self.request.user},
            )
            return redirect('backoffice:assembly-auth-app', assembly=self.assembly.pk, pk=app.pk)

        except Exception as err:
            messages.error(self.request, str(err))
            return redirect('backoffice:assembly-auth', assembly=self.assembly.pk)


class AuthGetTokenView(AssemblyMixin, View):
    assembly_url_param = 'assembly'
    assembly_management = True

    def post(self, *args, **kwargs):
        technical_user = self.assembly.create_technical_user_if_necessary()
        Token.objects.filter(user=technical_user).delete()
        token = Token(user=technical_user)
        token.save()

        messages.success(self.request, format_lazy(_('Assembly__authentication__newtoken'), assembly=self.assembly.name, token=token.key))
        logger.info(
            'New API auth token generated for "%(technical_user)s" (assembly %(assembly)s by %(user)s)',
            {'technical_user': technical_user, 'assembly': self.assembly, 'user': self.request.user},
        )
        return redirect('backoffice:assembly-auth', assembly=self.assembly.pk)


class AuthAppView(AssemblyMixin, UpdateView):
    model = Application
    fields = ['name', 'client_id', 'client_type', 'authorization_grant_type', 'redirect_uris', 'internal_description']
    readonly_fields = ['client_id', 'client_type', 'authorization_grant_type']

    template_name = 'backoffice/assembly_auth_app.html'
    assembly_url_param = 'assembly'
    assembly_management = True

    def get_form(self, *args, **kwargs):
        form = super().get_form(*args, **kwargs)
        for x in self.readonly_fields:
            form.fields[x].disabled = True
        return form

    def get_object(self, *args, **kwargs):
        obj = super().get_object(*args, **kwargs)
        if not Application.type_is(obj):  # pragma: no cover
            raise ValueError('Object is not an Application')
        if obj.assembly_id != self.assembly.id:
            raise Application.DoesNotExist
        return obj

    def form_valid(self, form):
        form.instance.save()
        messages.success(self.request, _('updated'))
        logger.info(
            'Updated OAuth2 app "%(app_name)s" for assembly %(assembly)s by %(user)s',
            {'app_name': form.instance.name, 'assembly': self.assembly, 'user': self.request.user},
        )
        return redirect('backoffice:assembly-auth-app', assembly=self.assembly.pk, pk=self.object.id)


class VouchersView(AssemblyMixin, TemplateView):
    template_name = 'backoffice/assembly_vouchers.html'

    assembly_url_param = 'pk'
    assembly_management = True

    def get_queryset(self):
        return Voucher.objects.for_assembly(self.assembly).order_by('name')

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        qs = self.get_queryset()

        vouchers = []
        for voucher in qs:
            v = {'id': voucher.id, 'data': voucher.data, 'name': voucher.name, 'description': voucher.description_html}

            if voucher.hide_from_staff and not self.assembly.user_can_manage(self.request.user, staff_can_manage=False):
                v['contents'] = None
            else:
                v['contents'] = voucher.entries.filter(assigned_assembly=self.assembly).values_list('content', flat=True)

            vouchers.append(v)

        ctx['vouchers'] = vouchers
        return ctx
