import logging
from typing import Any

from django.conf import settings
from django.contrib import messages
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import View
from django.views.generic.edit import CreateView, FormView, UpdateView

from core.integrations import BigBlueButton, Hangar, IntegrationError
from core.models.conference import ConferenceExportCache, ConferenceMember
from core.models.events import Event
from core.models.rooms import Room, RoomLink

from backoffice.forms import (
    AssemblyRoomCreateBigBlueButtonForm,
    AssemblyRoomCreateGenericForm,
    AssemblyRoomCreateHangarForm,
    AssemblyRoomEditForm,
    AssemblyRoomEditHangarForm,
    AssemblyRoomLinkCreateForm,
)
from backoffice.views.mixins import AssemblyMixin

logger = logging.getLogger(__name__)


class RoomNotAvailableError(Exception):
    pass


class AssemblyRoomUpdateView(AssemblyMixin, UpdateView):
    model = Room
    form_class = AssemblyRoomEditForm
    template_name = 'backoffice/assembly_room.html'

    assembly_url_param = 'assembly'
    create_roomlink_form = None

    def __init__(self, *args, link_edit=False, **kwargs):
        self.link_edit = link_edit
        super().__init__(*args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['createform'] = self.create_roomlink_form or AssemblyRoomLinkCreateForm()
        ctx['supports_links'] = self.object.room_type in Room.TYPES_WITH_LINKS
        ctx['can_delete'] = not Event.objects.filter(conference=self.conference, room=self.object).exists()
        ctx['link_edit'] = self.link_edit

        if self.object.room_type == Room.RoomType.HANGAR:
            ctx['legal_warning'] = _('Room-new-hangar__legal_hint')
            ctx['can_delete'] = False

        return ctx

    def get_object(self, *args, **kwargs):
        room = super().get_object(*args, **kwargs)
        if not Room.type_is(room):  # pragma: no cover
            raise ValueError('Object is not an Room')
        assert room.assembly == self.assembly
        return room

    def get_form_kwargs(self) -> dict[str, Any]:
        member = ConferenceMember.get_member(user=self.request.user, conference=self.conference)
        return {
            **super().get_form_kwargs(),
            'schedule_admin': member.has_perms('core.scheduleadmin', require_staff=True),
        }

    def get_form(self, *args, **kwargs):
        if self.object.room_type == Room.RoomType.HANGAR:
            # keeping the form readonly. To make it writeable, do `return EditAssemblyRoomHangarForm(**self.get_form_kwargs())``
            return AssemblyRoomEditHangarForm(instance=self.object)

        return super().get_form(*args, **kwargs)

    def form_valid(self, form):
        res = super().form_valid(form)
        form.instance.save()

        logger.info(
            'changed room "%(room)s" on assembly %(assembly)s by %(user)s', {'room': form.instance, 'assembly': self.assembly, 'user': self.request.user}
        )
        messages.success(self.request, _('updated'))
        ConferenceExportCache.signal_schedule_modification(self.conference, obj=self)

        return res

    def get_success_url(self):
        return reverse('backoffice:assembly-room', kwargs={'assembly': self.assembly.id, 'pk': self.object.id})


class AssemblyRoomCreateView(AssemblyMixin, FormView):
    assembly_url_param = 'assembly'
    template_name = 'backoffice/room-new.html'
    assembly_management = True

    def dispatch(self, *args, **kwargs):
        self.room_type = (self.request.POST if self.request.method != 'GET' else self.request.GET).get('type', None)
        try:
            return super().dispatch(*args, **kwargs)
        except RoomNotAvailableError:
            return redirect('backoffice:assembly-create-room', assembly=self.assembly.id)

    def get_form(self, *args, **kwargs):
        if self.room_type is None:
            return None

        if self.room_type == Room.RoomType.BIGBLUEBUTTON:
            if BigBlueButton is None or not BigBlueButton.can_create_for_assembly(self.assembly):
                messages.error(self.request, 'BBB not available')
                raise RoomNotAvailableError

            return AssemblyRoomCreateBigBlueButtonForm(self.request.POST, assembly=self.assembly)

        if self.room_type == Room.RoomType.HANGAR:
            if Hangar is None or not Hangar.can_create_for_assembly(self.assembly):
                messages.error(self.request, 'Hangar not available')
                raise RoomNotAvailableError

            return AssemblyRoomCreateHangarForm(self.request.POST, assembly=self.assembly)

        if self.room_type in Room.RoomType.values:
            return AssemblyRoomCreateGenericForm(self.request.POST, assembly=self.assembly, initial={'room_type': self.room_type})

        else:
            logger.warning('Unexpected room_type "%s" upon creating new room for %s.', self.room_type, self.assembly)
            messages.warning(self.request, _('internal_error_please_retry'))
            raise RoomNotAvailableError

    def get_context_data(self, *args, **kwargs):
        ctx = super().get_context_data(*args, **kwargs)
        ctx['room_type'] = self.room_type

        if not self.room_type:
            ctx['rooms_available'] = rooms_available = {k[0]: True for k in Room.RoomType.choices}
            rooms_available.update(
                {
                    'bbb': BigBlueButton is not None and BigBlueButton.can_create_for_assembly(self.assembly),
                    'hangar': Hangar is not None and Hangar.can_create_for_assembly(self.assembly),
                }
            )
            ctx['support_bbb'] = settings.INTEGRATIONS_BBB
            ctx['support_hangar'] = settings.INTEGRATIONS_HANGAR

        elif self.room_type == Room.RoomType.HANGAR:
            ctx['legal_warning'] = _('Room-new-hangar__legal_hint')

        return ctx

    def form_valid(self, form):
        try:
            room = form.create_room(self.request)
            if room:
                logger.info(
                    'Created %(room_type)s room "%(room)s" (assembly %(assembly)s by %(user)s)',
                    {'room_type': room.room_type, 'room': room.name, 'assembly': self.assembly, 'user': self.request.user},
                )
        except IntegrationError as err:
            messages.error(self.request, str(err))
            room = None

        if not room:
            return render(self.request, self.template_name, self.get_context_data())

        return redirect('backoffice:assembly-room', assembly=self.assembly.id, pk=form.room_id)


class RoomLinkCreateView(AssemblyMixin, CreateView):
    assembly_url_param = 'assembly'
    assembly_management = True
    model = RoomLink
    form_class = AssemblyRoomLinkCreateForm

    def get(self, *args, **kwargs):
        return redirect('backoffice:assembly-room', assembly=self.assembly.id, pk=kwargs.get('room'))

    def post(self, *args, **kwargs):
        self.room = self.assembly.rooms.get(id=kwargs.get('room'))
        return super().post(*args, **kwargs)

    def form_valid(self, form):
        room_link = form.instance
        room_link.room = self.room
        room_link.save()

        messages.success(self.request, format_lazy(_('RoomLink__created'), name=room_link.name))
        logger.info(
            'Link to %(link_type)s "%(roomlink)s" added to %(assembly)s room "%(room)s" by %(user)s: %(link)s',
            {
                'link': room_link.link,
                'link_type': room_link.link_type,
                'roomlink': room_link.name,
                'room': self.room.name,
                'assembly': self.assembly,
                'user': self.request.user,
            },
        )

        return redirect('backoffice:assembly-room', assembly=self.assembly.id, pk=self.room.id)

    def form_invalid(self, form):
        reason = ' '.join(' '.join(v) for k, v in form.errors.items())
        messages.warning(self.request, format_lazy('{info}: {reason}', info=_('RoomLink__create_failed'), reason=reason))

        # re-use the original room view
        view = AssemblyRoomUpdateView(request=self.request, assembly=self.assembly, link_edit=True)
        view.get_object = lambda: self.room
        view.create_roomlink_form = form
        return view.get(self.request)


class RoomLinkDeleteView(AssemblyMixin, View):
    assembly_url_param = 'assembly'
    assembly_management = True

    def post(self, *args, **kwargs):
        roomlink = RoomLink.objects.get(room__assembly_id=self.assembly.id, pk=self.request.POST.get('link'))
        roomlink.delete()

        messages.success(self.request, format_lazy(_('RoomLink__removed'), name=roomlink.name))
        return redirect('backoffice:assembly-room', assembly=self.assembly.id, pk=roomlink.room_id)


class AssemblyRoomDeleteView(AssemblyMixin, View):
    assembly_url_param = 'assembly'
    assembly_management = True

    def post(self, *args, **kwargs):
        room = self.assembly.rooms.get(pk=kwargs.get('room'))
        if Event.objects.filter(conference=self.conference, room=room).exclude(room__room_type=Room.RoomType.PROJECT).exists():
            messages.error(self.request, _('Room__remove__failed'))
            return redirect('backoffice:assembly-room', assembly=self.assembly.id, pk=room.id)

        room.delete()
        messages.success(self.request, _('Room__remove__success %s') % room.name)
        ConferenceExportCache.signal_schedule_modification(self.conference, obj=room)
        return redirect('backoffice:assembly', pk=self.assembly.id)
