import json
import logging
import re
from pathlib import Path
from typing import TypeIs
from uuid import uuid4

import rules
from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.gis.db import models as gis_models
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q, QuerySet
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from django.utils.html import escape as html_escape
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from rules.contrib.models import RulesModel
from rules.predicates import is_authenticated

from core.fields import ConferenceReference
from core.markdown import compile_translated_markdown_fields, render_markdown, store_relationships
from core.models.activitylog import ActivityLogEntry, ActivityLogMixin
from core.models.base_managers import ConferenceManagerMixin
from core.models.conference import Conference, ConferenceMember
from core.models.invitation import Invitation
from core.models.map import MapFloor
from core.models.tags import TaggedItemMixin, TagItem
from core.models.users import PlatformUser
from core.utils import render_markdown_as_text
from core.validators import FileSizeValidator, ImageDimensionValidator


@rules.predicate
def is_assembly_member(user: PlatformUser, assembly: 'Assembly') -> bool:
    if assembly is None:
        return False
    return assembly.has_user(user)


@rules.predicate
def is_assembly_manager(user: PlatformUser, assembly: 'Assembly') -> bool:
    if assembly is None:
        return False
    return assembly.user_can_manage(user)


@rules.predicate
def is_habitat_manager(user: PlatformUser, assembly: 'Assembly') -> bool:
    if assembly is None or assembly.parent is None:
        return False
    return assembly.parent.user_can_manage(user)


def perm_is(permission: str):
    @rules.predicate
    def user_has_permission(user, assembly: 'Assembly'):
        member = ConferenceMember.get_member(conference=assembly.conference, user=user)
        return member.has_perms(permission, require_all=True, require_staff=True)

    return user_has_permission


class AssemblyManager(ConferenceManagerMixin['Assembly']):
    staff_permissions = ['core.assembly_team']
    assembly_filter = 'self'

    def apply_public_filter(self, queryset: 'QuerySet[Assembly]', member: ConferenceMember | None = None) -> 'QuerySet[Assembly]':
        return queryset.filter(Q(state__in=self.model.PUBLIC_STATES))

    def associated_to_user(self, user: PlatformUser, conference: Conference) -> 'QuerySet[Assembly]':
        # guests cannot manage anything
        if user is None or not user.is_authenticated:
            return Assembly.objects.none()
        return self.associated_with_user(conference, user=user).filter(members__member=user).exclude(Q(state__in=[Assembly.State.HIDDEN]))


def get_banner_file_name(instance: 'Assembly', filename: str):
    return str(Path(str(instance.id)).joinpath('banner', filename))


class Assembly(TaggedItemMixin, ActivityLogMixin, RulesModel):
    class Meta:
        constraints = [
            models.constraints.UniqueConstraint(fields=['conference', 'slug'], name='unique_conferenceassembly_slug'),
        ]
        verbose_name = _('Assembly')
        verbose_name_plural = _('Assemblies')
        rules_permissions = {
            'view': is_assembly_member,
            'add': is_authenticated,
            'change': is_assembly_manager | perm_is('core.change_assembly'),
            'delete': is_assembly_manager | perm_is('core.delete_assembly'),
            'leave_habitat': is_assembly_manager | is_habitat_manager | perm_is('core.change_assembly'),
        }

    class State(models.TextChoices):
        PLANNED = 'planned', _('Assembly__state-planned')
        REGISTERED = 'registered', _('Assembly__state-registered')
        HIDDEN = 'hidden', _('Assembly__state-hidden')
        ACCEPTED = 'accepted', _('Assembly__state-accepted')
        REJECTED = 'rejected', _('Assembly__state-rejected')
        PLACED = 'placed', _('Assembly__state-placed')
        ARRIVED_SELF = 'arrived', _('Assembly__state-arrived_self')
        ARRIVED_CONFIRMED = 'confirmed', _('Assembly__state-arrived_confirmed')

    class Hierarchy(models.TextChoices):
        REGULAR = 'regular', _('Assembly__hierarchy-regular')
        CLUSTER = 'cluster', _('Assembly__hierarchy-cluster')
        CLUSTER_RESTRICTED = 'clusterrestricted', _('Assembly__hierarchy-clusterrestricted')

    objects = AssemblyManager()
    logger = logging.getLogger(__name__)

    id = models.UUIDField(default=uuid4, primary_key=True, editable=False)
    conference = ConferenceReference(related_name='assemblies')
    slug = models.SlugField(help_text=_('Assembly__slug__help'), verbose_name=_('Assembly__slug'))
    name = models.CharField(max_length=200, help_text=_('Assembly__name__help'), verbose_name=_('Assembly__name'))

    is_official = models.BooleanField(default=False, help_text=_('Assembly__is_official__help'), verbose_name=_('Assembly__is_official'))

    banner_image_height = models.PositiveIntegerField(blank=True, null=True)
    banner_image_width = models.PositiveIntegerField(blank=True, null=True)
    banner_image = models.ImageField(
        blank=True,
        null=True,
        height_field='banner_image_height',
        width_field='banner_image_width',
        help_text=_('Assembly__banner_image__help %(limit)d, %(max_width)d, %(max_height)d')
        % {
            'limit': settings.ASSEMBLY_BANNER_FILE_SIZE_LIMIT,
            'max_width': settings.ASSEMBLY_BANNER_WIDTH_MAXIMUM,
            'max_height': settings.ASSEMBLY_BANNER_HEIGHT_MAXIMUM,
        },
        upload_to=get_banner_file_name,
        verbose_name=_('Assembly__banner_image'),
        validators=[
            FileSizeValidator(settings.ASSEMBLY_BANNER_FILE_SIZE_LIMIT),
            ImageDimensionValidator(
                min_size=(
                    settings.ASSEMBLY_BANNER_WIDTH_MINIMUM,
                    settings.ASSEMBLY_BANNER_HEIGHT_MINIMUM,
                ),
                max_size=(
                    settings.ASSEMBLY_BANNER_WIDTH_MAXIMUM,
                    settings.ASSEMBLY_BANNER_HEIGHT_MAXIMUM,
                ),
            ),
        ],
    )

    description = models.TextField(blank=True, null=True, help_text=_('Assembly__description__help'), verbose_name=_('Assembly__description'))
    description_html = models.TextField(blank=True, null=True)
    registration_details = models.TextField(
        blank=True, null=True, help_text=_('Assembly__registration_details__help'), verbose_name=_('Assembly__registration_details')
    )

    registration_data = models.JSONField(blank=True, null=True, help_text=_('Assembly__registration_data__help'), verbose_name=_('Assembly__registration_data'))

    state = models.CharField(
        max_length=20, default=State.PLANNED, choices=State.choices, verbose_name=_('Assembly__state'), help_text=_('Assembly__state__help')
    )
    hierarchy = models.CharField(
        max_length=20, default=Hierarchy.REGULAR, choices=Hierarchy.choices, help_text=_('Assembly__hierarchy__help'), verbose_name=_('Assembly__hierarchy')
    )
    parent = models.ForeignKey(
        'self', blank=True, null=True, related_name='+', on_delete=models.PROTECT, verbose_name=_('Assembly__parent'), help_text=_('Assembly__parent__help')
    )

    technical_user = models.ForeignKey(
        PlatformUser,
        blank=True,
        null=True,
        related_name='+',
        on_delete=models.PROTECT,
        verbose_name=_('Assembly__technical_user'),
        help_text=_('Assembly__technical_user__help'),
    )

    is_virtual = models.BooleanField(blank=True, null=True, verbose_name=_('Assembly__is_virtual'), help_text=_('Assembly__is_virtual__help'))
    is_physical = models.BooleanField(blank=True, null=True, verbose_name=_('Assembly__is_physical'), help_text=_('Assembly__is_physical__help'))
    is_remote = models.BooleanField(blank=True, null=True, verbose_name=_('Assembly__is_remote'), help_text=_('Assembly__is_remote__help'))

    assembly_location = models.TextField(blank=True, null=True, verbose_name=_('Assembly__location'), help_text=_('Assembly__location__help'))
    assembly_link = models.URLField(blank=True, null=True, verbose_name=_('Assembly__link'), help_text=_('Assembly__link__help'))
    public_contact = models.EmailField(
        blank=True,
        null=True,
        help_text=_('Assembly__public_contact__help'),
        verbose_name=_('Assembly__public_contact'),
    )

    sent_invitations = GenericRelation(
        Invitation,
        content_type_field='requester_type',
        object_id_field='requester_id',
        related_query_name='assemblies_inviters',
    )
    received_invitations = GenericRelation(
        Invitation,
        content_type_field='requested_type',
        object_id_field='requested_id',
        related_query_name='assemblies_invited',
    )

    location_floor = models.ForeignKey(
        MapFloor,
        blank=True,
        null=True,
        related_name='assemblies',
        on_delete=models.PROTECT,
        help_text=_('Assembly__location_floor__help'),
        verbose_name=_('Assembly__location_floor'),
    )

    location_point = gis_models.PointField(
        blank=True,
        null=True,
        help_text=_('Assembly__location_point__help'),
        verbose_name=_('Assembly__location_point'),
    )

    location_boundaries = gis_models.MultiPolygonField(
        blank=True,
        null=True,
        help_text=_('Assembly__location_boundaries__help'),
        verbose_name=_('Assembly__location_boundaries'),
    )

    location_data = models.JSONField(
        blank=True,
        null=True,
    )
    """temporary work-around field for c3nav x/y coordinate scheme"""

    created = models.DateTimeField(
        auto_now_add=True,
        help_text=_('Assembly__created__help'),
        verbose_name=_('Assembly__created'),
    )
    last_update_assembly = models.DateTimeField(
        blank=True,
        null=True,
        help_text=_('Assembly__last_update_assembly__help'),
        verbose_name=_('Assembly__last_update_assembly'),
    )
    last_update_staff = models.DateTimeField(
        blank=True,
        null=True,
        help_text=_('Assembly__last_update_staff__help'),
        verbose_name=_('Assembly__last_update_staff'),
    )
    activity_log_entries = GenericRelation(
        ActivityLogEntry,
        object_id_field='entity_object_id',
        content_type_field='entity_content_type',
    )

    favorite_of = models.ManyToManyField(
        PlatformUser, related_name='favorite_assemblies', help_text=_('Assembly__favorite_of__help'), verbose_name=_('Assembly__favorite_of')
    )
    tags = GenericRelation(TagItem, object_id_field='target_id', content_type_field='target_type', related_query_name='assemblies')

    PUBLIC_STATES = [
        State.ACCEPTED.value,
        State.PLACED.value,
        State.ARRIVED_CONFIRMED.value,
        State.ARRIVED_SELF.value,
    ]

    PLACED_STATES = [
        State.PLACED.value,
        State.ARRIVED_CONFIRMED.value,
        State.ARRIVED_SELF.value,
    ]

    def __str__(self):
        return self.name

    @classmethod
    def type_is(cls, obj: object) -> TypeIs['Assembly']:
        return isinstance(obj, cls)

    def get_absolute_url(self):
        from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

        return hub_absolute('plainui:assembly', assembly_slug=self.slug)

    def ensure_single_type_conference_match(self) -> dict[str, tuple[bool, bool]]:
        """
        Checks if the conference only supports a single type of assembly
        and sets that field if it is not set.
        """
        changes = {}

        ssat = self.conference.get_single_supported_assembly_type()
        if ssat is not None:
            if ssat == 'physical' and not self.is_physical:
                changes['is_physical'] = (self.is_physical, True)
                self.is_physical = True
            elif ssat == 'virtual' and not self.is_virtual:
                changes['is_virtual'] = (self.is_virtual, True)
                self.is_virtual = True
            elif ssat == 'remote' and not self.is_remote:
                changes['is_remote'] = (self.is_remote, True)
                self.is_remote = True

        return changes

    def clean(self):
        self.slug = self.slug.strip() if self.slug is not None else None
        self.name = self.name.strip() if self.name is not None else None
        super().clean()

        errors = {}

        self.ensure_single_type_conference_match()

        # an assembly must be at least one of physical/virtual/remote (among those supported by the conference)
        set_physical = self.conference.support_assembly_physical and self.is_physical
        set_virtual = self.conference.support_assembly_virtual and self.is_virtual
        set_remote = self.conference.support_assembly_remote and self.is_remote
        if self.state in self.PUBLIC_STATES and not (set_physical or set_virtual or set_remote):
            x = _('Assembly__need_type')
            if self.conference.support_assembly_physical:
                errors['is_physical'] = x
            if self.conference.support_assembly_remote:
                errors['is_remote'] = x
            if self.conference.support_assembly_virtual:
                errors['is_virtual'] = x

        # ensure the technical user has type 'assembly'
        if self.technical_user is not None:
            if self.technical_user.user_type != PlatformUser.Type.ASSEMBLY:
                errors['technical_user'] = _('Assembly__technical_user__must_be_assembly')

        # ensure the slug is not a forbidden one
        if self.slug in settings.FORBIDDEN_ASSEMBLY_SLUGS:
            errors['slug'] = _('Assembly__slug__is_forbidden')

        if errors:
            raise ValidationError(errors)

    @cached_property
    def conference_slug(self):
        return self.conference.slug

    @cached_property
    def description_html_short(self):
        if not self.description:
            return ''

        part = re.split(r'\r?\n\r?\n', self.description or '')[0]
        escaped = html_escape(part)
        escaped = re.sub(r'\*([\w\d() ]+)\*', lambda x: format_html('<b>{}</b>', x.group(1)), escaped)
        escaped = re.sub(r'__([\w\d() ]+)__', lambda x: format_html('<i>{}</i>', x.group(1)), escaped)

        return mark_safe(escaped)

    @cached_property
    def is_public(self):
        return self.state in self.PUBLIC_STATES

    @cached_property
    def is_placed(self):
        return self.state in self.PLACED_STATES

    @cached_property
    def is_cluster(self):
        return self.hierarchy in [self.Hierarchy.CLUSTER.value, self.Hierarchy.CLUSTER_RESTRICTED.value]

    @cached_property
    def is_public_cluster(self):
        return self.is_public and self.hierarchy in [self.Hierarchy.CLUSTER.value]

    @property
    def children(self):
        return Assembly.objects.filter(parent=self).order_by('name')

    @property
    def public_children(self):
        return Assembly.objects.filter(parent=self).filter(state__in=Assembly.PUBLIC_STATES).order_by('name')

    @property
    def linked_assemblies(self):
        qs = self.assembly_links.filter(is_public=True, type__in=AssemblyLink.PUBLIC_TYPES)
        qs = qs.order_by('b__name')
        return [x.b for x in qs.select_related('b')]

    @property
    def base_data_readonly(self):
        return self.state not in [
            Assembly.State.PLANNED,
            Assembly.State.REGISTERED,
        ]

    def get_all_orga_contacts(self):
        return PlatformUser.objects.filter(pk__in=self.members.filter(is_representative=True).values_list('member_id', flat=True))

    def get_all_managing_contacts(self):
        return PlatformUser.objects.filter(pk__in=self.members.filter(can_manage=True).values_list('member_id', flat=True))

    def create_technical_user_if_necessary(self):
        if self.technical_user is not None:
            return self.technical_user

        username = f'_{self.conference_slug}__{self.slug}'
        try:
            u = PlatformUser(user_type=PlatformUser.Type.ASSEMBLY, username=username, is_active=True)
            u.save()
            self.logger.info('Created technical user "%s" for assembly <%s> (%s).', username, self.slug, self.id)

        except Exception as err:
            self.logger.error('Failed to create a technical user "%s" for assembly <%s> (%s): %s', username, self.slug, self.id, err)
            raise Exception('Technical user for assembly could not be created.')

        cm = u.conferences.create(conference=self.conference)
        cm.save()

        self.technical_user = u
        self.save()
        return u

    def has_user(self, user: PlatformUser):
        """Checks if the supplied user is listed as a member (guest, regular, manager or representative) of this assembly."""
        if not user.is_authenticated:
            return False

        return self.members.filter(member=user).exists()

    def user_can_manage(self, user: PlatformUser, staff_can_manage: bool = False):
        if not user.is_authenticated:
            return False

        if staff_can_manage and user.has_conference_staff_permission(self.conference, 'core.assembly_team'):
            return True

        return self.members.filter(member=user, can_manage_assembly=True).exists()

    def save(self, *args, update_fields=None, **kwargs):
        if update_fields is None or 'description' in update_fields:
            render_results = compile_translated_markdown_fields(self, self.conference, 'description')
            store_relationships(self.conference, self, render_results)

        if update_fields is None or 'slug' in update_fields:
            if self.slug in settings.FORBIDDEN_ASSEMBLY_SLUGS:
                raise Exception("Won't save assembly with forbidden slug!")

        return super().save(*args, update_fields=update_fields, **kwargs)

    def prepare_mail_to_managers(self, subject: str, message: str, *, include_assembly_name: bool = True) -> tuple[str, str | None, str]:
        from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

        ctx = {
            'conference': self.conference,
            'assembly': self,
            'subject': subject,
            'message': render_markdown_as_text(message) if settings.SUPPORT_HTML_MAILS else message,
            'public_link': self.get_absolute_url(),
            'backoffice_link': hub_absolute('backoffice:assembly-edit', pk=self.pk),
        }

        body_text = render_to_string('core/assembly_notification.txt', ctx)

        if settings.SUPPORT_HTML_MAILS:
            ctx['message_html'] = render_markdown(conf=self.conference, markup=message, dont_derefer_allowlist=True)
            body_html = render_to_string('core/assembly_notification.html', ctx)
        else:
            body_html = None

        msg_subj = f'[{self.conference.slug}]'
        if include_assembly_name:
            msg_subj += f' {{{self.name}}}'
        msg_subj += f' {subject}'

        return msg_subj, body_html, body_text

    def send_mail_to_managers(self, subject, message, prepare_only: bool = False) -> list[str]:
        msg_subj, body_html, body_text = self.prepare_mail_to_managers(subject, message)
        result = []
        for recipient in self.members.filter(can_manage_assembly=True):  # type: PlatformUser
            recipient.member.notify_user(
                subject=msg_subj,
                message_text=body_text,
                message_html=body_html,
                details_link=None,
                use_all_verified_channels=True,
                from_email=settings.ASSEMBLY_TEAM_FROM_EMAIL,
                reply_to=settings.ASSEMBLY_TEAM_REPLY_TO,
            )
            result.append(recipient.member.username)
        return result

    def get_voucher_count(self, with_always_public: bool = True) -> int | None:
        from .voucher import Voucher

        entries = Voucher.objects.for_assembly(self, include_public_ones=False).count()

        # if we shall not pay respect to Vouchers available to everyone, just return the number here
        if not with_always_public:
            return entries

        # if there are entries, we have our result
        if entries > 0:
            return entries

        # check if there are "public" vouchers, return integer in that case or None otherwise (no available vouchers)
        has_for_all_assemblies = Voucher.objects.available_for_conference(self.conference).exists()
        return 0 if has_for_all_assemblies else None

    def get_location_floor_index(self):
        return self.location_floor.index if self.location_floor else None

    def get_location_point_xy(self):
        """Get the location_point's X/Y coordinate as array of (two) floats, if it's a valid coordinate - return None otherwise."""
        if not self.location_point or not self.location_point.valid:
            return None

        return [self.location_point.x, self.location_point.y]

    def get_location_point_as_json(self):
        return json.dumps(self.get_location_point_xy() or '')

    def get_location_boundaries_xy(self):
        if not self.location_boundaries or not self.location_boundaries.valid:
            return None

        polygons = []
        for multipolygon in self.location_boundaries:
            for polygon in multipolygon:
                polygons.append([[x[0], x[1]] for x in polygon.array])
        return polygons

    def get_location_boundaries_as_json(self):
        return json.dumps(self.get_location_boundaries_xy() or '')

    def get_activity_log_kind_for_user(self, user: PlatformUser) -> ActivityLogEntry.Kind:
        if self.has_user(user):
            return ActivityLogEntry.Kind.ENTITY
        member = ConferenceMember.get_member(conference=self.conference, user=user)
        if member.has_perms('core.assembly_team', require_staff=True):
            return ActivityLogEntry.Kind.STAFF
        return ActivityLogEntry.Kind.OTHER


class AssemblyLinkManager(ConferenceManagerMixin['AssemblyLink']):
    staff_permissions = ['core.assembly_team']
    conference_filter = 'a__conference'

    def apply_public_filter(self, queryset: 'QuerySet[AssemblyLink]', member: ConferenceMember | None = None) -> 'QuerySet[AssemblyLink]':
        return queryset.filter(Type__in=self.model.PUBLIC_TYPES)


class AssemblyLink(models.Model):
    class Type(models.TextChoices):
        RELATED = 'related', _('AssemblyLink__type-related')
        """Both assemblies have a related topic/theme."""

        ASSOCIATED = 'associated', _('AssemblyLink__type-associated')
        """Both assemblies are interconnected with each other."""

        OTHER = 'other', _('AssemblyLink__type-other')

    a = models.ForeignKey(Assembly, on_delete=models.CASCADE, related_name='assembly_links')
    b = models.ForeignKey(Assembly, on_delete=models.CASCADE, related_name='related_assemblies')
    type = models.CharField(max_length=20, default=Type.RELATED, choices=Type.choices)

    is_public = models.BooleanField(default=False, verbose_name=_('AssemblyLink__is_public'), help_text=_('AssemblyLink__is_public__help'))

    objects = AssemblyLinkManager()

    PUBLIC_TYPES = [
        Type.RELATED.value,
        Type.ASSOCIATED.value,
    ]

    def clean(self, *args, **kwargs):
        if self.a.conference != self.b.conference:
            raise ValidationError(_('AssemblyLink__conference_mismatch_error'))

        return super().clean(*args, **kwargs)

    def __str__(self):
        s = f'{self.type}: {self.a.name} -> {self.b.name}'
        if not self.is_public:
            s = '/* ' + s + '*/'
        return s


class AssemblyMemberManager(ConferenceManagerMixin['AssemblyMember']):
    staff_permissions = ['core.assembly_team']
    conference_filter = 'assembly__conference'
    assembly_filter = 'assembly'

    def apply_public_filter(self, queryset: 'QuerySet[AssemblyMember]', member: ConferenceMember | None = None) -> 'QuerySet[AssemblyMember]':
        return queryset.filter(show_public=True)

    def manageable_by_user_for_assembly(self, *, user: PlatformUser, assembly: 'Assembly') -> 'QuerySet[AssemblyMember]':
        return self.manageable_by_user(conference=assembly.conference, user=user).filter(assembly=assembly)


class AssemblyMember(models.Model):
    objects = AssemblyMemberManager()

    assembly = models.ForeignKey(Assembly, related_name='members', on_delete=models.CASCADE)
    member = models.ForeignKey(PlatformUser, related_name='associated_assemblies', on_delete=models.PROTECT)

    is_representative = models.BooleanField(
        default=False,
        help_text=_('AssemblyMember__is_representative__help'),
        verbose_name=_('AssemblyMember__is_representative'),
    )

    can_manage_assembly = models.BooleanField(
        default=False,
        help_text=_('AssemblyMember__can_manage_assembly__help'),
        verbose_name=_('AssemblyMember__can_manage_assembly'),
    )

    is_content_coordinator = models.BooleanField(
        default=False,
        help_text=_('AssemblyMember__is_content_coordinator__help'),
        verbose_name=_('AssemblyMember__is_content_coordinator'),
    )

    is_production_coordinator = models.BooleanField(
        default=False,
        help_text=_('AssemblyMember__is_production_coordinator__help'),
        verbose_name=_('AssemblyMember__is_production_coordinator'),
    )

    is_technical_contact = models.BooleanField(
        default=False,
        help_text=_('AssemblyMember__is_technical_contact__help'),
        verbose_name=_('AssemblyMember__is_technical_contact'),
    )

    show_public = models.BooleanField(default=True, help_text=_('AssemblyMember__show_public__help'), verbose_name=_('AssemblyMember__show_public'))

    @property
    def roles(self):
        if self.is_representative:
            yield 'representative'
        if self.can_manage_assembly:
            yield 'manager'
        if self.is_content_coordinator:
            yield 'content_coordinator'
        if self.is_production_coordinator:
            yield 'production_coordinator'
        if self.is_technical_contact:
            yield 'technical_contact'

    def get_roles_display(self):
        res = []
        if self.is_representative:
            res.append(gettext('AssemblyMember__is_representative__help'))
        if self.can_manage_assembly:
            res.append(gettext('AssemblyMember__can_manage_assembly'))
        if self.is_content_coordinator:
            res.append(gettext('AssemblyMember__is_content_coordinator'))
        if self.is_production_coordinator:
            res.append(gettext('AssemblyMember__is_production_coordinator'))
        if self.is_technical_contact:
            res.append(gettext('AssemblyMember__is_technical_contact'))
        return ', '.join(res)

    def __str__(self):
        return f'{self.member} ({self.get_roles_display()})'
