import logging
import typing
import xml.etree.ElementTree as ET
from collections import OrderedDict
from datetime import date, datetime, time, timedelta
from datetime import timezone as tz
from json import JSONEncoder
from uuid import uuid4

from timezone_field import TimeZoneField

from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.postgres import fields as pg_fields
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator, validate_slug
from django.db import models
from django.db.models import Max
from django.http import HttpRequest, HttpResponse, HttpResponseNotModified
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _

from core.choices import UserBadgeVisibility
from core.fields import ConferenceReference
from core.models.users import PlatformUser
from core.utils import render_markdown_as_text

logger = logging.getLogger(__name__)


class ConferenceConfigurationException(RuntimeError):
    pass


class NoStartDateException(ConferenceConfigurationException):
    def __init__(self, message=None):
        self.message = message or 'Conference has no start date.'


class NoEndDateException(ConferenceConfigurationException):
    def __init__(self, message=None):
        self.message = message or 'Conference has no end date.'


class PrettyJSONEncoder(JSONEncoder):
    def __init__(self, *args, indent, sort_keys, **kwargs):
        super().__init__(*args, indent=2, sort_keys=True, **kwargs)


class ConferenceMember(models.Model):
    conference = ConferenceReference(related_name='users', on_delete=models.CASCADE)
    user = models.ForeignKey(PlatformUser, related_name='conferences', on_delete=models.CASCADE)

    uuid = models.UUIDField(default=uuid4, unique=True)

    active_angel = models.BooleanField(default=False)
    roles = pg_fields.ArrayField(models.CharField(max_length=50), blank=True, null=True)
    is_staff = models.BooleanField(default=False)
    has_ticket = models.BooleanField(default=False, help_text=_('ConferenceMember--has_ticket--help'), verbose_name=_('ConferenceMember--has_ticket'))

    permission_groups = models.ManyToManyField(Group, blank=True, related_name='+')
    static_page_groups = pg_fields.ArrayField(models.CharField(max_length=50), blank=True, null=True)

    description = models.TextField(blank=True, null=True, help_text=_('PlatformUser__description__help'), verbose_name=_('PlatformUser__description'))
    description_html = models.TextField(blank=True, null=True)

    default_badge_visibility = models.CharField(
        max_length=50,
        choices=UserBadgeVisibility.choices,
        default=UserBadgeVisibility.PRIVATE,
        help_text=_('ConferenceMember__default_badge_visibility__help'),
        verbose_name=_('ConferenceMember__default_badge_visibility'),
    )

    class Meta:
        permissions = [
            ('conference_admin', _('ConferenceMember__permission-confernece_admin')),
            # Change conference settings (e.g., dates, ), update venue json.
            ('assembly_team', _('ConferenceMember__permission-assembly_team')),
            # See all assemblies, not only the accepted ones.
            ('assembly_registration_admin', _('ConferenceMember__permission-assembly_registration_admin')),
            # See all assemblies, not only the accepted ones.
            ('static_pages', _('ConferenceMember__permission-static_pages')),
            # Access to static pages, can be further limited by configuring static_page_groups.
            ('map_edit', _('ConferenceMember__permission-map_edit')),
            # modification of the map, i.e. POIs
            ('moderation', 'Orga: Moderation'),
            # moderation of all content
            ('voucher_admin', _('ConferenceMember__permission-voucher_admin')),
            # configure conference's vouchers
            ('scheduleadmin', _('ConferenceMember__permission-scheduleadmin')),
            # manage/support ScheduleSupports globally, i.e. without access to an individual assembly
        ]

        constraints = [models.constraints.UniqueConstraint(fields=['user', 'conference'], name='conferencemember_user_conference_unique')]

    def get_permission_groups_permissions(self):
        """Returns only the permission codenames which are given via the associated permission_groups of this ConferenceMember instance."""

        my_permissions = {'.'.join(x) for x in self.permission_groups.values_list('permissions__content_type__app_label', 'permissions__codename')}
        return my_permissions

    @cached_property
    def all_permissions(self):
        """Returns all permission codenames relevant to this conference member: global ones and those specific to this conference member assignment."""
        all_permissions = self.user.get_all_permissions() | self.get_permission_groups_permissions()
        return all_permissions

    @property
    def is_authenticated(self):
        return self.user.is_authenticated

    @classmethod
    def get_member(
        cls,
        conference: 'Conference',
        *,
        user: PlatformUser | None = None,
        member: 'ConferenceMember | None' = None,
        prefetch_user: bool = False,
    ) -> 'ConferenceMember':
        if member is None and user is None:
            raise ValueError('Either user or member must be given.')
        if member and user and member.user != user:
            raise ValueError('User and member do not match.')
        if member and member.conference != conference:
            return cls.get_anonymous_member(conference, user)
        if member:
            return member
        if not user.is_authenticated:
            return cls.get_anonymous_member(conference, user)
        try:
            qs = cls.objects.all()
            if prefetch_user:
                qs = qs.select_related('user').prefetch_related('user__groups', 'user__user_permissions', 'permission_groups')
            return qs.get(conference=conference, user=user)
        except cls.DoesNotExist:
            if conference.require_ticket:
                return cls.get_anonymous_member(conference, user)
            logger.trace(
                'Creating new member for user "%(user)s" at conference "%(conference)s".',
                {
                    'user': user.username,
                    'conference': conference.slug,
                },
            )
            return cls.objects.create(conference=conference, user=user)

    @classmethod
    def get_anonymous_member(cls, conf: 'Conference', user: PlatformUser | None = None):
        logger.trace('Creating anonymous member for conference %s.', conf.slug)

        class AnonMember:
            def __init__(self):
                for x in cls._meta.fields:
                    setattr(self, x.name, x.default if x.default != models.fields.NOT_PROVIDED else None)

                self.user = PlatformUser.get_anonymous_user() if user is None else user
                self.roles = []

            def has_perms(self, *perms: str, require_all: bool = True, require_staff: bool = False) -> bool:
                return False

            conference = conf

            def save(self, *args, **kwargs):
                raise TypeError('Cannot save an anonymous member.')

        return AnonMember()

    def has_perms(self, *perms: str, require_all: bool = True, require_staff: bool = False) -> bool:
        """Check if a user has a set of permissions.

        Args:
            *perms (str): The list of permissions to check.
            require_all (bool, optional): Defines, if all permission are required or just one. Defaults to True.
            require_staff (bool, optional): If a staff status is also required. Defaults to False.

        Returns:
            bool: True if the user has all permissions.
        """
        if require_staff and not self.is_staff:
            return False
        if require_all:
            return all(perm in self.all_permissions for perm in perms)
        return any(perm in self.all_permissions for perm in perms)

    def __str__(self):
        return f'{self.user.username}@{self.conference.slug}'

    def save(self, *args, update_fields=None, **kwargs) -> None:
        from core.markdown import compile_translated_markdown_fields, store_relationships

        if update_fields is None or 'description' in update_fields:
            render_results = compile_translated_markdown_fields(self, self.conference, 'description')
            store_relationships(self.conference, self, render_results)

        return super().save(*args, update_fields=update_fields, **kwargs)


class ConferenceManager(models.Manager):
    def accessible_by_user(self, user: PlatformUser):
        # everybody gets the public conferences
        qs = self.get_queryset()

        # guests only get the public ones
        if user is None or not user.is_authenticated:
            return qs.filter(is_public=True)

        # for global staff apply no more limits
        if user.is_superuser or user.is_staff:
            return qs

        # limit on the public ones
        public_query = qs.filter(is_public=True).values('id')

        # for logged-in members, add all non-public conferences in which the user has the staff flag set
        conf_staff_query = ConferenceMember.objects.filter(user=user, is_staff=True, conference__is_public=False).values('conference_id')

        return qs.filter(id__in=public_query.union(conf_staff_query))

    def public_accessible(self):
        return self.get_queryset().filter(is_public=True)

    @cached_property
    def current_conference(self):
        """Returns the currently running public Conference."""
        qs = self.get_queryset().filter(is_public=True)

        # attempt: a Conference where now() is between start and end
        try:
            return qs.get(start__lte=timezone.now(), end__gte=timezone.now())
        except Exception:
            pass

        # attempt: a Conference with start in the past and open end
        try:
            return qs.get(start__lte=timezone.now(), end=None)
        except Exception:
            pass

        # attempt: first Conference starting in the future
        try:
            return qs.filter(start__gte=timezone.now()).first()
        except Exception:
            pass

        # all attempts failed, give up
        return None


class Conference(models.Model):
    class Meta:
        verbose_name = _('Conference')
        verbose_name_plural = _('Conferences')

    objects = ConferenceManager()

    id = models.UUIDField(default=uuid4, primary_key=True, editable=False)
    slug = models.SlugField(unique=True, help_text=_('Conference__slug__help'), verbose_name=_('Conference__slug'))
    name = models.CharField(max_length=200, help_text=_('Conference__name__help'), verbose_name=_('Conference__name'))

    is_public = models.BooleanField(default=False, help_text=_('Conference__is_public__help'), verbose_name=_('Conference__is_public'))
    publication_date = models.DateTimeField(
        blank=True, null=True, help_text=_('Conference__publication_date__help'), verbose_name=_('Conference__publication_date')
    )
    registration_start = models.DateTimeField(
        blank=True, null=True, help_text=_('Conference__registration_start__help'), verbose_name=_('Conference__registration_start')
    )
    registration_deadline = models.DateTimeField(
        blank=True, null=True, help_text=_('Conference__registration_deadline__help'), verbose_name=_('Conference__registration_deadline')
    )

    start = models.DateTimeField(blank=True, null=True, help_text=_('Conference__start__help'), verbose_name=_('Conference__start'))
    end = models.DateTimeField(blank=True, null=True, help_text=_('Conference__end__help'), verbose_name=_('Conference__end'))

    timezone = TimeZoneField(
        choices_display='WITH_GMT_OFFSET', default='UTC', help_text=_('Conference__timezone__help'), verbose_name=_('Conference__timezone')
    )
    global_notification = models.TextField(
        default='', blank=True, help_text=_('Conference__global_notification__help'), verbose_name=_('Conference__global_notification')
    )

    logo = models.TextField(
        blank=True,
        null=True,
        help_text=_('Conference__logo__help'),
        verbose_name=_('Conference__logo'),
    )
    """SVG logo"""

    send_pn_disabled = models.BooleanField(default=False, help_text=_('Conference__send_pn_disabled__help'), verbose_name=_('Conference__send_pn_disabled'))
    board_disabled = models.BooleanField(default=False, help_text=_('Conference__board_disabled__help'), verbose_name=_('Conference__board_disabled'))
    require_login = models.BooleanField(default=False, help_text=_('Conference--require_login--help'), verbose_name=_('Conference--require_login'))
    require_ticket = models.BooleanField(default=False, help_text=_('Conference--require_ticket--help'), verbose_name=_('Conference--require_ticket'))

    support_clusters = models.BooleanField(default=True, help_text=_('Conference__support_clusters__help'), verbose_name=_('Conference__support_clusters'))
    additional_fields_schema = models.JSONField(
        blank=True,
        null=True,
        encoder=PrettyJSONEncoder,
        help_text=_('Conference__additional_fields_schema__help'),
        verbose_name=_('Conference__additional_fields_schema'),
    )

    disclaimers = models.JSONField(
        blank=True, null=True, encoder=PrettyJSONEncoder, help_text=_('Conference__disclaimers__help'), verbose_name=_('Conference__disclaimers')
    )

    self_organized_sessions_assembly = models.ForeignKey(
        'Assembly',
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name='+',
        help_text=_('Conference__self_organized_sessions_assembly__help'),
        verbose_name=_('Conference__self_organized_sessions_assembly'),
    )
    """
    Benutzer, die eine Self-Organized-Session (SOS) anlegen wollen,
    können dies entweder in ihrer Assembly tun, in denen sie "Member" sind oder aber in dieser hier.
    """

    support_assembly_physical = models.BooleanField(
        default=True,
        help_text=_('Conference__support_assembly_physical__help'),
        verbose_name=_('Conference__support_assembly_physical'),
    )

    support_assembly_virtual = models.BooleanField(
        default=True,
        help_text=_('Conference__support_assembly_virtual__help'),
        verbose_name=_('Conference__support_assembly_virtual'),
    )

    support_assembly_remote = models.BooleanField(
        default=True,
        help_text=_('Conference__support_assembly_remote__help'),
        verbose_name=_('Conference__support_assembly_remote'),
    )

    mail_footer = models.TextField(
        blank=True,
        null=True,
        help_text=_('Conference__mail_footer__help'),
        verbose_name=_('Conference__mail_footer'),
    )

    map_config = models.JSONField(
        blank=True,
        null=True,
        encoder=PrettyJSONEncoder,
        help_text=_('Conference__map_config__help'),
        verbose_name=_('Conference__map_config'),
    )
    """
    Konfiguration für die eingebundene Karten wie z. B. Typ, Startposition/-bereich, verfügbare/angezeigte Layer, etc.
    """

    def clean(self):
        super().clean()

        errors = {}

        # verify logo is an XML-parseable SVG file
        if self.logo:
            try:
                parsed = ET.fromstring(self.logo)
                assert parsed.tag in ['svg', '{http://www.w3.org/2000/svg}svg'], 'Not a SVG document.'

                assert 'width' not in parsed and 'height' not in parsed, 'SVG has width and/or height attribute.'

                xml_start = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>'
                if self.logo.startswith(xml_start):
                    self.logo = self.logo[len(xml_start) :].lstrip()

            except ET.ParseError:
                errors['logo'] = 'SVG (XML) parse error'
            except AssertionError as err:
                errors['logo'] = str(err)

        # at least one assembly type must be supported
        if not (self.support_assembly_physical or self.support_assembly_virtual or self.support_assembly_remote):
            x = _('Conference__support_assembly_type')
            errors['support_assembly_physical'] = x
            errors['support_assembly_remote'] = x
            errors['support_assembly_virtual'] = x

        if errors:
            raise ValidationError(errors)

    @cached_property
    def is_open(self):
        return (
            self.is_public
            and (self.registration_start is None or self.registration_start < timezone.now())
            and (self.registration_deadline is None or self.registration_deadline > timezone.now())
        )

    @cached_property
    def is_running(self):
        if self.start is None:
            return False
        # A conference may have no end date, in which case it is considered running indefinitely.
        if self.end is None:
            return self.start <= timezone.now()
        return self.start <= timezone.now() <= self.end

    @cached_property
    def is_published(self):
        return self.is_public and (self.publication_date is None or self.publication_date < timezone.now())

    @cached_property
    def has_ended(self):
        return self.end is not None and timezone.now() > self.end

    @cached_property
    def days_count(self):
        if self.start is None:
            return 0
        end = self.end if self.end is not None else timezone.now().replace(hour=23, minute=59, second=59)
        return (end - self.start).days + 1

    @cached_property
    def days(self):
        days = []
        if self.start is None:
            return days
        end = self.end if self.end is not None else timezone.now().replace(hour=23, minute=59, second=59)
        for i in range(self.days_count):
            date = self.start.date() + timedelta(days=i)
            days.append(ConferenceDay(i, date, self.timezone))

        days[0].start = self.start
        days[-1].end = end
        return days

    @cached_property
    def _day0(self) -> datetime:
        if self.start is None:
            raise NoStartDateException
        return self.start.replace(hour=6, minute=0, second=0) - timedelta(days=1)

    def get_conference_day(self, timestamp: datetime) -> int:
        """
        Returns a numeric conference day for the presented timestamp.
        Adheres to the CCC scheme of turning the day at 6am on-site time (instead of midnight).
        If no start is defined the day will be 0.
        """
        try:
            return (timestamp - self._day0).days
        except NoStartDateException:
            return 0

    @property
    def current_day(self):
        return self.get_conference_day(timezone.now())

    @cached_property
    def get_additional_fields_schema(self):
        schema = self.additional_fields_schema
        if schema is None:
            return None

        for v in schema:
            if 'choices' in v:
                ordered_choices = OrderedDict()
                for ck in sorted(v['choices'], key=lambda y: f"{v['choices'][y].get('_index', 1000):04}-{y}"):
                    ordered_choices[ck] = v['choices'][ck]
                v['choices'] = ordered_choices

        return schema

    @cached_property
    def field_names(self):
        result = set()
        for f in self.additional_fields_schema or []:
            result.add(f.get('key'))
        return sorted(result)

    def get_field(self, name):
        for f in self.additional_fields_schema or []:
            if f.get('key') == name:
                return f
        return None

    def has_disclaimer(self, key):
        """Checks if a disclaimer text is present for the given key."""
        exists = self.disclaimers is not None and key in self.disclaimers
        return exists

    def get_disclaimer(self, key, lang=None):
        """
        Returns the disclaimer text with the given key.
        If lang is given, extract the corresponding dict key, otherwise return all translations.

        @returns None if the key does not exist, '' or the translated value depending on whether that can be found
        """
        value = self.disclaimers.get(key, None)
        if value is None:
            return None

        if not isinstance(value, dict):
            return value

        if lang is None:
            lang = get_language()

        for x in [lang, 'de', 'en']:
            v = value.get(x)
            if v is not None:
                return v

        logging.warning('No disclaimer for lang "%s" on conference "%s".', lang, self.slug)
        return ''

    def get_single_supported_assembly_type(self):
        """Query the single supported assembly type, if any. Returns the type if a single supported one exists or None otherwise."""
        if self.support_assembly_physical and not self.support_assembly_virtual and not self.support_assembly_remote:
            return 'physical'

        if not self.support_assembly_physical and self.support_assembly_virtual and not self.support_assembly_remote:
            return 'virtual'

        if not self.support_assembly_physical and not self.support_assembly_virtual and self.support_assembly_remote:
            return 'remote'

        return None

    def get_navigation_tree(self):
        """
        Retrieves the ConferenceNavigationItems ready to be used in e.g. a template
        (only visible items, in correct order, URLs are resolved).
        """
        result = []

        # prefetch a list of all children (avoid a DB roundtrip per parent/category)
        all_children = list(self.nav_items.exclude(parent=None).filter(is_visible=True).order_by('index'))

        # iterate over the categories (parent entries)
        for parent in self.nav_items.filter(parent=None, is_visible=True).order_by('index'):
            # filter children for this parent entry (they're already sorted) and resolve the URL, if necessary
            this_children = [
                {
                    'label': c.label,
                    'title': c.title,
                    'icon': c.icon,
                    'url': c.resolve_url(),
                }
                for c in all_children
                if c.parent_id == parent.id
            ]

            # append parent entry, reformatted as dict
            result.append(
                {
                    'label': parent.label,
                    'title': parent.title,
                    'icon': parent.icon,
                    'children': this_children,
                }
            )

        return result

    @cached_property
    def logo_html(self):
        if not self.logo:
            with settings.BASE_DIR.joinpath('hub', 'logos', 'default.svg').open() as default_logo:
                default_logo_content = default_logo.read()
            assert default_logo_content.startswith('<svg'), 'Expected default logo to start with SVG tag.'
            return mark_safe(default_logo_content)

        return mark_safe(self.logo)

    @cached_property
    def mail_footer_text(self):
        if not self.mail_footer:
            return None

        return '\n-- \n' + render_markdown_as_text(self.mail_footer)

    @cached_property
    def mail_footer_html(self):
        if not self.mail_footer:
            return None

        from core.markdown import render_markdown

        return format_html(
            '<div id="footer" style="margin-top: 1em; border-top: 2px solid #999; color: #999; font-size: 80%;">{0}</div>',
            render_markdown(conf=self, markup=self.mail_footer, dont_derefer_allowlist=True),
        )

    def __str__(self):
        return self.name


class ConferenceTrackManager(models.Manager):
    def accessible_by_user(self, user: PlatformUser, conference: Conference):
        assert user is not None

        qs = self.get_queryset()
        if conference is not None:
            qs = qs.filter(conference=conference)

        # guests only get the public ones
        if user is None or not user.is_authenticated:
            return qs.filter(is_public=True)

        # for global staff apply no more limits
        if user.is_staff or user.is_superuser:
            return qs

        # see if we can return all tracks (i.e. the user is marked as "staff")
        if conference is not None:
            try:
                member = ConferenceMember.objects.get(conference=conference, user=user)
                if member.is_staff:
                    return qs
            except ConferenceMember.DoesNotExist:
                pass

        # for everybody else, only show tracks marked public
        return qs.filter(is_public=True)

    def public_accessible(self):
        return self.get_queryset().filter(is_public=True)


class ConferenceTrack(models.Model):
    class Meta:
        constraints = [
            models.constraints.UniqueConstraint(fields=['conference', 'slug'], name='unique_conferencetrack_slug'),
        ]
        verbose_name = _('ConferenceTrack')
        verbose_name_plural = _('ConferenceTracks')

    objects = ConferenceTrackManager()

    conference = ConferenceReference(related_name='tracks')
    slug = models.SlugField(help_text=_('ConferenceTrack__slug__help'), verbose_name=_('ConferenceTrack__slug'))
    name = models.CharField(max_length=200, help_text=_('ConferenceTrack__name__help'), verbose_name=_('ConferenceTrack__name'))
    color = models.CharField(blank=True, max_length=7, help_text=_('ConferenceTrack__color__help'), verbose_name=_('ConferenceTrack__color'))
    is_public = models.BooleanField(default=True, help_text=_('ConferenceTrack__is_public__help'), verbose_name=_('ConferenceTrack__is_public'))
    banner_image_height = models.PositiveIntegerField(blank=True, null=True)
    banner_image_width = models.PositiveIntegerField(blank=True, null=True)
    banner_image = models.ImageField(
        blank=True,
        null=True,
        height_field='banner_image_height',
        width_field='banner_image_width',
        help_text=_('ConferenceTrack__banner_image__help'),
        verbose_name=_('ConferenceTrack__banner_image'),
    )

    last_update = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name


class ConferenceDay:
    index: int
    start: datetime
    end: datetime

    def __init__(self, i: int, date: date, time_zone: tz):
        self.index = i + 1
        self.start = datetime.combine(date, time(6, 0)).astimezone(time_zone)
        self.end = datetime.combine(date + timedelta(days=1), time(6, 0)).astimezone(time_zone)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, ConferenceDay):
            return False
        return self.index == value.index and self.start == value.start and self.end == value.end

    def __repr__(self):
        return f'{self.index} ({self.start.date()} - {self.end.date()})'


class ConferenceExportCache(models.Model):
    class Type(models.TextChoices):
        SCHEDULE = 'schedule'
        MAP = 'map'
        ASSEMBLIES = 'assemblies'
        LOGO = 'logo'

    conference = ConferenceReference(related_name='+')
    type = models.CharField(max_length=20, choices=Type.choices)
    ident = models.CharField(max_length=150)
    etag = models.CharField(max_length=100, editable=False)
    needs_regeneration = models.BooleanField(default=True)
    last_generated = models.DateTimeField(blank=True, null=True)
    data = models.TextField(blank=True, null=True)

    class Meta:
        unique_together = [
            ('conference', 'ident'),
        ]

    @property
    def is_dirty(self):
        # we're flagged for regeneration already or have no data -> yes, we're dirty
        if self.needs_regeneration is True or self.data is None or self.last_generated is None:
            return True

        # check per-type if updates are available
        if self.type == self.Type.SCHEDULE:
            from .events import Event
            from .rooms import Room

            # get latest update timestamp for ConferenceTracks, Events and Rooms
            last_track = ConferenceTrack.objects.filter(conference=self.conference).aggregate(Max('last_update'))['last_update__max']
            last_event = Event.objects.filter(conference=self.conference).aggregate(Max('last_update'))['last_update__max']
            last_room = Room.objects.filter(conference=self.conference).aggregate(Max('last_update'))['last_update__max']

            # if one of those is greater than our last generation, we need to update
            if (
                (last_track is not None and last_track > self.last_generated)
                or (last_event is not None and last_event > self.last_generated)
                or (last_room is not None and last_room > self.last_generated)
            ):
                return True

        elif self.type == self.Type.MAP:
            from .assemblies import Assembly

            last_assemblies = Assembly.objects.filter(conference=self.conference).aggregate(Max('last_update_staff'))['last_update_staff__max']
            if last_assemblies is not None and last_assemblies > self.last_generated:
                return True

        elif self.type == self.Type.ASSEMBLIES:
            from .assemblies import Assembly

            last_updates = Assembly.objects.filter(conference=self.conference).aggregate(Max('last_update_assembly'), Max('last_update_staff'))
            last_assemblies = last_updates['last_update_assembly__max']
            last_staff = last_updates['last_update_staff__max']
            if (last_assemblies is not None and last_assemblies > self.last_generated) or (last_staff is not None and last_staff > self.last_generated):
                return True

        else:
            raise NotImplementedError('Unexpected value for ConferenceExportCache.type!')

        # no need to update found yet -> there is none
        return False

    def save(self, *args, **kwargs):
        from core.utils import int_to_custom_string

        if self.last_generated is not None:
            self.etag = int_to_custom_string(int(self.last_generated.timestamp() * 1000))
        return super().save(*args, **kwargs)

    @classmethod
    def _get_or_create_entry(cls, conference: Conference, entry_type: Type, ident: str, result_func: typing.Callable[..., bytes]):
        cache_entry, _ = cls.objects.get_or_create(
            conference=conference,
            type=entry_type,
            ident=ident,
            defaults={'needs_regeneration': True},
        )
        if cache_entry.is_dirty or settings.DEBUG:
            cache_entry.data = result_func()
            cache_entry.last_generated = timezone.now()
            cache_entry.needs_regeneration = False
            cache_entry.save()
        return cache_entry

    @classmethod
    def fetch_entry(cls, conference: Conference, entry_type: Type, ident: str, result_func, as_text: bool = False):
        cache_entry = cls._get_or_create_entry(conference, entry_type, ident, result_func)
        content = cache_entry.data
        return content.decode('utf-8') if as_text else content

    @classmethod
    def handle_http_request(cls, request: HttpRequest, conference: Conference, entry_type: Type, ident: str, content_type, result_func):
        cache_entry = cls._get_or_create_entry(conference, entry_type, ident, result_func)

        headers = {
            'Content-Type': content_type if isinstance(content_type, str) else content_type(),
            'ETag': cache_entry.etag,
            'Last-Modified': cache_entry.last_generated.isoformat(),
        }

        content = cache_entry.data.encode('utf-8')

        if 'If-None-Match' in request.headers:
            if request.headers['If-None-Match'] == headers['ETag']:
                return HttpResponseNotModified(headers=headers)

        if 'If-Modified-Since' in request.headers:
            try:
                ts_req = datetime.fromisoformat(request.headers['If-Modified-Since'])
                ts_resp = datetime.fromisoformat(headers['Last-Modified'])
                if ts_req >= ts_resp:
                    return HttpResponseNotModified(headers=headers)
            except ValueError:
                # ignore timestamp parsing error, handle as non-match
                pass

        return HttpResponse(content, headers=headers)

    @classmethod
    def signal_schedule_modification(cls, conference, obj=None):
        """
        Signals that the conference's schedule got an update and will need to be regenerated.
        :type conference: Conference
        :type obj: Assembly|Event|Room
        """
        # TODO: check if modified obj is relevant to the cached schedule(s)
        cls.objects.filter(conference=conference, type=cls.Type.SCHEDULE).update(needs_regeneration=True)

    @classmethod
    def signal_assembly_modification(cls, conference, assembly=None):
        """
        Signals that the conference's assemblies got an update and exports may need to be regenerated.
        :type conference: Conference
        :type assembly: Assembly
        """
        # TODO: check if modified assembly is relevant to the cached schedule(s)
        cls.objects.filter(conference=conference, type__in=[cls.Type.ASSEMBLIES, cls.Type.MAP]).update(needs_regeneration=True)


def validate_conferencenavigationitem_url(value: str) -> None:
    if '://' not in value:
        raise ValidationError(_('ConferenceNavigationItem__url__missingseparator'))

    _valid_protocols = ['assembly', 'event', 'conference', 'wiki']

    protocol, rhs = value.split('://', maxsplit=1)
    if protocol in _valid_protocols:
        # ignore optional query string
        rhs_split = rhs.split('?', maxsplit=1)
        validate_slug(rhs_split[0])
    elif protocol in ['http', 'https']:
        URLValidator()(value)
    else:
        raise ValidationError(_('ConferenceNavigationItem__url__supportedprotocols') + 'http, https, ' + ', '.join(_valid_protocols))


class ConferenceNavigationItem(models.Model):
    class Meta:
        verbose_name = _('ConferenceNavigationItem')
        verbose_name_plural = _('ConferenceNavigationItems')

    conference = ConferenceReference(related_name='nav_items')
    parent = models.ForeignKey(
        'self',
        related_name='children',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text=_('ConferenceNavigationItem__parent__help'),
        verbose_name=_('ConferenceNavigationItem__parent'),
    )

    is_visible = models.BooleanField(
        default=True,
        help_text=_('ConferenceNavigationItem__is_visible__help'),
        verbose_name=_('ConferenceNavigationItem__is_visible'),
    )
    index = models.IntegerField(
        help_text=_('ConferenceNavigationItem__index__help'),
        verbose_name=_('ConferenceNavigationItem__index'),
    )
    icon = models.CharField(
        max_length=50,
        blank=True,
        null=True,
        help_text=_('ConferenceNavigationItem__icon__help'),
        verbose_name=_('ConferenceNavigationItem__icon'),
    )

    label = models.CharField(
        max_length=50,
        help_text=_('ConferenceNavigationItem__label__help'),
        verbose_name=_('ConferenceNavigationItem__label'),
    )
    title = models.CharField(
        max_length=200,
        help_text=_('ConferenceNavigationItem__title__help'),
        verbose_name=_('ConferenceNavigationItem__title'),
    )
    url = models.CharField(
        max_length=255,
        blank=True,
        validators=[validate_conferencenavigationitem_url],
        help_text=_('ConferenceNavigationItem__url__help'),
        verbose_name=_('ConferenceNavigationItem__url'),
    )

    def resolve_url(self):
        from core.utils import resolve_internal_url

        return resolve_internal_url(self.url, fallback_as_is=True)

    def __str__(self):
        return f'{self.parent} > {self.label}' if self.parent is not None else self.label
