from urllib.parse import urlparse
from xml.etree import ElementTree

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference


class MetaNavItemManager(models.Manager):
    def for_conference(self, conf):
        return self.get_queryset().filter(conference=conf, visible=True).order_by('index')


class MetaNavItem(models.Model):
    class Meta:
        verbose_name = _('MetaNavItem')
        verbose_name_plural = _('MetaNavItems')

    objects = MetaNavItemManager()

    conference = ConferenceReference(related_name='meta_nav_items')

    visible = models.BooleanField(
        default=True,
        help_text=_('MetaNavItem__visible__help'),
        verbose_name=_('MetaNavItem__visible'),
    )
    """Shall the item be displayed at all?"""
    enabled = models.BooleanField(
        default=True,
        help_text=_('MetaNavItem__enabled__help'),
        verbose_name=_('MetaNavItem__enabled'),
    )
    """Only relevant if visible=True: shall the item be clickable or only be there as a placeholder?"""

    index = models.IntegerField(
        help_text=_('MetaNavItem__index__help'),
        verbose_name=_('MetaNavItem__index'),
    )
    slug = models.SlugField(
        unique=True,
        help_text=_('MetaNavItem__slug__help'),
        verbose_name=_('MetaNavItem__slug'),
    )
    title = models.CharField(
        max_length=100,
        help_text=_('MetaNavItem__title__help'),
        verbose_name=_('MetaNavItem__title'),
    )

    graphic_light = models.TextField()
    """Holds the (UTF-8 encoded) SVG for use with a light theme."""
    graphic_dark = models.TextField(blank=True, null=True)
    """Holds the (UTF-8 encoded) SVG for use with a dark theme. If left blank, the light version will be used."""

    url = models.URLField(
        blank=True,
        help_text=_('MetaNavItem__url__help'),
        verbose_name=_('MetaNavItem__url'),
    )

    def __str__(self):
        return f'[{self.slug}] {self.title}' + (' (HIDDEN)' if not self.visible else '') + (' (DISABLED)' if not self.enabled else '')

    def get_graphic_light_as_html(self):
        return mark_safe(self.graphic_light)

    def get_graphic_dark(self, default_to_light: bool = True, for_template: bool = False):
        img = self.graphic_dark
        if not img and default_to_light:
            img = self.graphic_light
        return mark_safe(img) if for_template else img

    def get_graphic_dark_as_html(self, default_to_light: bool = True):
        return self.get_graphic_dark(default_to_light=default_to_light, for_template=True)

    def clean(self):
        super().clean()
        errors = {}

        if self.url:
            parsed_url = urlparse(self.url)
            if not parsed_url.netloc:
                errors['url'] = _('MetaNavItem__url__requiredabsolute')
        elif self.visible and self.enabled:
            errors['url'] = _('MetaNavItem__url__necessary_when_shown')

        if not self.graphic_light:
            errors['graphic_light'] = _('MetaNavItem__graphic_light__required')

        for fld in ['graphic_light', 'graphic_dark']:
            v = getattr(self, fld)
            if not v:
                continue

            try:
                decoded = ElementTree.fromstring(v)
                if decoded.tag not in ['svg', '{http://www.w3.org/2000/svg}svg']:
                    errors[fld] = _('MetaNavItem__graphic_not_a_svg')
            except (ValueError, ElementTree.ParseError):
                errors[fld] = _('MetaNavItem__graphic_not_a_svg')

        if errors:
            raise ValidationError(errors)
