import logging
from uuid import uuid4

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference
from core.markdown import render_markdown
from core.models.assemblies import Assembly
from core.models.conference import Conference
from core.models.users import PlatformUser


class VoucherManager(models.Manager):
    def available_for_conference(self, conference: Conference):
        assert conference is not None

        # filter on enabled vouchers in the given conference
        qs = self.get_queryset().filter(conference=conference, enabled=True)
        return qs

    def for_assembly(self, assembly: Assembly, include_public_ones: bool = True):
        assert assembly is not None

        # pre-filter on enabled vouchers in the assembly's conference
        qs = self.get_queryset().filter(conference=assembly.conference, enabled=True)

        # filter on the correct target type
        qs = qs.filter(target=Voucher.Target.ASSEMBLY)

        # select vouchers which shall always be shown or ones assigned to the assembly
        associated_voucher_ids = VoucherEntry.objects.filter(assigned_assembly=assembly).distinct('voucher_id').values_list('voucher_id', flat=True)
        q = Q(id__in=associated_voucher_ids)
        if include_public_ones:
            q |= Q(show_always=True)
        qs = qs.filter(q)

        # we're done
        return qs


class Voucher(models.Model):
    class Meta:
        verbose_name = _('Voucher')
        verbose_name_plural = _('Vouchers')

    class Assignment(models.TextChoices):
        MANUAL = 'manual', _('Voucher__assignment-manual')
        """Manual (individual) assignment of Voucher only."""

        AUTO = 'automatic', _('Voucher__assignment-automatic')
        """"""

        ON_PUBLIC = 'on_public', _('Voucher__assignment-on_public')
        """Assign when the assembly is made public."""

    class Data(models.TextChoices):
        TEXT = 'text', _('Voucher__data-text')
        """Just some string to be shown to the users."""

        URL = 'url', _('Voucher__data-url')
        """An URL which will be clickable by the user."""

    class Target(models.TextChoices):
        ASSEMBLY = 'assembly', _('Voucher__target-assembly')
        """This voucher is for assemblies."""

        USER = 'user', _('Voucher__target-user')
        """This voucher is for an user."""

    id = models.UUIDField(default=uuid4, primary_key=True, editable=False)
    conference = ConferenceReference(related_name='vouchers')
    name = models.CharField(max_length=200, help_text=_('Voucher__name__help'), verbose_name=_('Voucher__name'))

    enabled = models.BooleanField(default=False, help_text=_('Voucher__enabled__help'), verbose_name=_('Voucher__enabled'))
    """Mark the voucher as being available. This is intended to be used as a 'draft' marker."""

    show_always = models.BooleanField(default=False, help_text=_('Voucher__show_always__help'), verbose_name=_('Voucher__show_always'))
    """Show the voucher in any case for a target, even if it hasn't been assigned yet."""

    hide_from_staff = models.BooleanField(default=False, help_text=_('Voucher__hide_from_staff__help'), verbose_name=_('Voucher__hide_from_staff'))
    """Hide the values even from staff (unless they are Voucher Admin)."""

    assignment = models.CharField(
        max_length=20,
        blank=False,
        choices=Assignment.choices,
        default=Assignment.MANUAL,
        help_text=_('Voucher__assignment__help'),
        verbose_name=_('Voucher__assignment'),
    )

    data = models.CharField(
        max_length=20, blank=False, choices=Data.choices, default=Data.TEXT, help_text=_('Voucher__data__help'), verbose_name=_('Voucher__data')
    )

    target = models.CharField(max_length=20, blank=False, choices=Target.choices, help_text=_('Voucher__target__help'), verbose_name=_('Voucher__target'))

    description = models.TextField(blank=True, null=True, help_text=_('Voucher__description__help'), verbose_name=_('Voucher__description'))

    objects = VoucherManager()
    logger = logging.getLogger(__name__)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.full_clean()  # ensure field errors are catched _before_ saving
        super().save(*args, **kwargs)

    @cached_property
    def description_html(self):
        if self.description is None:
            return ''

        html = render_markdown(self.conference, self.description)
        return html

    @classmethod
    def do_auto_assignments(cls, conference: Conference = None, assemblies: list[Assembly] | None = None) -> int:
        qs = cls.objects.filter(conference=conference) if conference is not None else cls.objects.all()
        qs = qs.filter(enabled=True)
        qs = qs.filter(target=Voucher.Target.ASSEMBLY)

        total = 0
        for voucher in qs.iterator():
            if conference is None and assemblies is not None:
                v_assemblies = [a for a in assemblies if a.conference_id == conference.pk]
            else:
                v_assemblies = assemblies
            total += voucher.do_auto_assignment(v_assemblies)

        return total

    def do_auto_assignment(self, assemblies: list[Assembly]) -> int:
        if self.target != self.Target.ASSEMBLY:
            raise NotImplementedError('Auto-Assignment of non assemblies not implemented yet.')

        # nothing to do if we're not enabled or set to manual assignment
        if (not self.enabled) or self.assignment == self.Assignment.MANUAL:
            return 0

        # work on all conference assemblies if no list is provided
        if assemblies is None:
            if self.assignment == self.Assignment.ON_PUBLIC:
                assemblies = self.conference.assemblies.conference_accessible(self.conference)
            elif self.assignment == self.Assignment.AUTO:
                assemblies = self.conference.assemblies.all()
            else:
                raise NotImplementedError('Voucher: unexpected value of .assignment in .do_auto_assignment(None)')

        # fetch list of assemblies not already assigned
        assigned_assembly_ids = self.entries.exclude(assigned_assembly=None).values_list('assigned_assembly_id', flat=True)
        missing_assignments = [a for a in assemblies if a.id not in assigned_assembly_ids]

        # fetch available entries
        available_entries = list(self.entries.filter(assigned=None).all())

        # iterate over all missing assignments
        total = 0
        for assembly in missing_assignments:
            # skip non-public assembly if auto-assignment is for public ones only
            if self.assignment == self.Assignment.ON_PUBLIC:
                if self.target == self.Target.ASSEMBLY and not assembly.is_public:
                    continue

            # abort (but warn) if no more available entries are available
            if not available_entries:
                self.logger.warning('Not enough free entries for auto-assignment of voucher %s.', self.id)
                break

            # match the next available entry with this assembly
            entry = available_entries.pop()
            entry.assign(assembly)
            self.logger.info('Auto-assigned voucher %s to assembly <%s> (%s, entry #%s).', self.id, assembly.slug, assembly.id, entry.id)

            total += 1

        return total


class VoucherEntry(models.Model):
    voucher = models.ForeignKey(Voucher, on_delete=models.CASCADE, related_name='entries')
    content = models.TextField(blank=True, null=True, help_text=_('VoucherEntry__content__help'), verbose_name=_('VoucherEntry__content'))

    created = models.DateTimeField(auto_now_add=True, help_text=_('VoucherEntry__created__help'), verbose_name=_('VoucherEntry__created'))
    expires = models.DateTimeField(blank=True, null=True, help_text=_('VoucherEntry__expires__help'), verbose_name=_('VoucherEntry__expires'))
    assigned = models.DateTimeField(blank=True, null=True, help_text=_('VoucherEntry__assigned__help'), verbose_name=_('VoucherEntry__assigned'))

    assigned_assembly = models.ForeignKey(Assembly, on_delete=models.SET_NULL, related_name='vouchers', blank=True, null=True)
    assigned_user = models.ForeignKey(PlatformUser, on_delete=models.SET_NULL, related_name='vouchers', blank=True, null=True)

    def __str__(self):
        return f'{self.voucher.name}#{self.id}'

    @cached_property
    def target(self):
        if self.voucher.target == Voucher.Target.ASSEMBLY:
            return self.assigned_assembly

        if self.voucher.target == Voucher.Target.USER:
            return self.assigned_user

        raise ValueError(f'Unknown voucher target: {self.voucher.target}')

    def clean(self):
        super().clean()

        errors = {}

        if self.assigned is not None:
            if self.assigned_assembly is None and self.assigned_user is None:
                errors['assigned'] = _('VoucherEntry__assigned__needtarget')

        else:
            # assembly assigned without a timestamp?
            if self.voucher.target == Voucher.Target.ASSEMBLY and self.assigned_assembly is not None:
                self.assigned = timezone.now()
            else:
                self.assigned_assembly = None

            # user assigned without a timestamp?
            if self.voucher.target == Voucher.Target.USER and self.assigned_user is not None:
                self.assigned = timezone.now()
            else:
                self.assigned_user = None

        if errors:
            raise ValidationError(errors)

    def save(self, *args, **kwargs):
        self.full_clean()  # ensure field errors are catched _before_ saving
        super().save(*args, **kwargs)

    def assign(self, target):
        if target is None:
            raise ValueError('VoucherEntry assignment target must not be null.')

        if self.assigned is not None or self.assigned_assembly is not None or self.assigned_user is not None:
            raise ValidationError(_('VoucherEntry__already_assigned'))

        if self.voucher.target == Voucher.Target.ASSEMBLY:
            if not isinstance(target, Assembly):
                raise ValueError("VoucherEntry's target must be an assembly.")

            if target.conference_id != self.voucher.conference_id:
                raise ValueError('Voucher and assembly must be in the same conference!')

            self.assigned = timezone.now()
            self.assigned_assembly = target
            self.assigned_user = None
            self.save(update_fields=['assigned', 'assigned_assembly', 'assigned_user'])

        elif self.voucher.target == Voucher.Target.USER:
            if not isinstance(target, PlatformUser):
                raise ValueError("VoucherEntry's target must be an user.")

            if self.voucher.conference.require_ticket and not target.is_in_conference(self.voucher.conference):
                raise ValueError('User must be in the conference.')

            self.assigned = timezone.now()
            self.assigned_assembly = None
            self.assigned_user = target
            self.save(update_fields=['assigned', 'assigned_assembly', 'assigned_user'])

        else:
            raise NotImplementedError("Unexpected value of Voucher.target, don't know how to assign!")
