import logging
from functools import lru_cache

import jwt
from oauth2_provider.scopes import BaseScopes

from django.conf import settings
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _

from .models.conference import Conference
from .models.users import PlatformUser

logger = logging.getLogger(__name__)


class SingleSignOn:
    def __init__(self):
        self._secret = settings.SSO_SECRET
        settings.SSO_SECRET = 'removed'

    def generate_token(self, user: PlatformUser):
        if user is None or not user.is_active:
            raise Exception('Inactive user.')

        data = {'uid': user.pk}
        return jwt.encode(data, self._secret, algorithm='HS256')

    def get_user_from_token(self, token: str) -> PlatformUser:
        try:
            data = jwt.decode(token, self._secret, algoriths=['HS256'])
            user = PlatformUser.objects.get(pk=data['uid'], is_active=True)
            return user

        except Exception as err:
            logger.error('Failed to get user from token: %s', err)
            return None


SSO = SingleSignOn()


class OAuth2Scopes(BaseScopes):
    @lru_cache
    def get_all_scopes(self):
        """
        Return a dict-like object with all the scopes available in the
        system. The key should be the scope name and the value should be
        the description.

        ex: {"read": "A read scope", "write": "A write scope"}
        """
        scopes = {
            'introspection': 'Token Introspection',
        }
        for c in Conference.objects.values('slug', 'name'):
            scopes[c['slug'] + '_attendee'] = format_lazy(_('SSO__conf_attendee'), **c)
            scopes[c['slug'] + '_service_angels'] = format_lazy(_('SSO__svc_angels'), **c)
            scopes[c['slug'] + '_service_usermgmt'] = format_lazy(_('SSO__svc_usermgmt'), **c)

        return scopes

    @lru_cache
    def _get_available_scopes(self, application):
        if application is None:
            return []

        conf_slug = application.assembly.conference.slug
        user_type = application.user.user_type

        result = []
        if user_type == PlatformUser.Type.ASSEMBLY:
            result.append(conf_slug + '_attendee')
            result.extend(k for k in self.get_all_scopes() if k.startswith(conf_slug + '_user_'))

        elif user_type == PlatformUser.Type.SERVICE:
            result.extend(k for k in self.get_all_scopes() if k.startswith(conf_slug + '_service_'))

        else:
            # nobody else should have an Application anyway
            logger.warning('OAuth2Scopes.get_available_scopes() called for unexpected application user_type: %s', application.id)
        return result

    def get_available_scopes(self, *args, application=None, request=None, **kwargs):
        """
        Return a list of scopes available for the current application/request.

        ex: ["read", "write"]
        """
        available_scopes = self._get_available_scopes(application=application)
        return available_scopes

    @lru_cache
    def _get_default_scopes(self, application):
        if application is None:
            return []

        conf_slug = application.assembly.conference.slug

        if application.user.user_type == PlatformUser.Type.ASSEMBLY:
            return [conf_slug + '_user_name']

        return []

    def get_default_scopes(self, *args, application=None, request=None, **kwargs):
        """
        Return a list of the default scopes for the current application/request.
        This MUST be a subset of the scopes returned by `get_available_scopes`.

        ex: ["read"]
        """
        if application is None or application.assembly is None:
            logger.warning('OAuth2Scopes.get_default_scopes() called without valid application in request: %s', request)
            return []

        return [
            application.assembly.conference.slug + '_attendee',
        ]
