import logging
import random
import re
import shutil
import subprocess
import tempfile
import uuid
from datetime import UTC, datetime, timedelta
from io import BytesIO
from pathlib import Path
from string import ascii_letters, digits
from urllib.parse import parse_qs, urlparse, urlunparse

import requests

from django.urls import NoReverseMatch
from django.utils.functional import cached_property
from django.utils.html import strip_tags

logger = logging.getLogger(__name__)


def scheme_and_netloc_from_url(url: str) -> str | None:
    parsed = urlparse(url)
    if not parsed.netloc:
        return None

    scheme = parsed.scheme or 'https'
    return scheme + '://' + parsed.netloc


def url_in_allowlist(scheme_and_netloc: str, patterns: list[re.Pattern | str]) -> bool:
    """
    Checks whether the given scheme_and_netloc is present in the given list of patterns.
    This is used in dereferrer handling and (optionally) markdown generation.
    """
    for pattern in patterns:
        if isinstance(pattern, re.Pattern):
            if pattern.match(scheme_and_netloc):
                return True
        elif scheme_and_netloc == pattern:
            return True

    return False


RE_MARKDOWN_LINKS = re.compile(r'\[([^\]]+)\]\(([^)]+)\)')


def render_markdown_as_text(markup: str):
    """
    Renders the given markdown snippet as plain text.
    This can be used in e.g. mails to provide a plain text part.
    """
    text_only = strip_tags(markup)
    link_urls = {}
    for link_text, link_url in RE_MARKDOWN_LINKS.findall(markup):
        if link_url not in link_urls:
            link_urls[link_url] = len(link_urls) + 1
        idx = link_urls[link_url]
        reference = f'{link_text}[{idx}]'
        text_only = text_only.replace(f'[{link_text}]({link_url})', reference)

    url_footer = '\n'.join([f'[{idx}] {link_url}' for link_url, idx in link_urls.items()])

    return text_only + (('\n\n' + url_footer) if url_footer else '')


def generate_token(length: int = 50, char_set: str = ''.join([ascii_letters, digits])):
    """
    Generates a random token string with given length.
    """
    rand = random.SystemRandom()
    token = ''.join([rand.choice(char_set) for i in range(length)])
    return token


def int_to_custom_string(number: int, alphabet: str = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    if alphabet is None or len(alphabet) < 2:
        raise ValueError('Alphabet must consist of at least two characters.')

    base = len(alphabet)
    result = []

    while number > 0:
        number, remainder = divmod(number, base)
        result.append(alphabet[remainder])

    return ''.join(result[::-1]) if result else alphabet[0]


def str2bool(s, allow_null: bool = False):
    if isinstance(s, bool):
        return s

    s = s.strip().lower()

    if allow_null and s in ['', '-', 'none', 'null', 'unknown', '?']:
        return None

    if s in ['true', 'on', '1', 't', 'y', 'yes', 'j', 'ja', '✓', '✔']:
        return True

    if s in ['false', 'off', '0', 'f', 'n', 'no', 'nein', '✗', '✘']:
        return False

    raise ValueError('Cannot parse as boolean: ' + s)


_RE_STR2TIMEDELTA = re.compile(r'((?P<hours>\d+?)hr?\s*)?((?P<minutes>\d+?)m(ins?)?\s*)?((?P<seconds>\d+?)s)?')


def str2timedelta(s):
    if ':' in s:
        parts = s.split(':')
        kwargs = {'seconds': int(parts.pop())}
        if parts:
            kwargs['minutes'] = int(parts.pop())
        if parts:
            kwargs['hours'] = int(parts.pop())
        if parts:
            kwargs['days'] = int(parts.pop())
        return timedelta(**kwargs)

    parts = _RE_STR2TIMEDELTA.match(s)
    if not parts:
        return
    parts = parts.groupdict()
    time_params = {}
    for name, param in parts.items():
        if param:
            time_params[name] = int(param)
    return timedelta(**time_params)


def unix2timestamp(seconds_since_epoch):
    """Converts seconds since EPOCH into a timezone-aware timestamp."""
    return datetime.fromtimestamp(seconds_since_epoch, tz=UTC)


def mask_url(url):
    """
    Take an URL and strip possibly sensitive parts from it,
    especially query parameters and credentials.

    As an example, <https://johndoe:secret@example.com/path?a=b>
    becomes <https://example.com/path>.
    """
    u = urlparse(url)
    loc = u.hostname + (f':{u.port}' if u.port is not None else '') if u.hostname is not None else u.netloc
    masked = (u.scheme, loc, u.path, u.params, None, None)
    return urlunparse(masked)


def resolve_internal_url(url: str, fallback_as_is: bool = True) -> str | None:
    """
    Resolves special URLs like
     - conference://url_resolver_name
     - assembly://slug
     - event://slug
     - wiki://slug

    Regular URLs (i.e. https://...) are returned as-is by default but may be resolved as None (i.e. no match).
    """
    try:
        protocol, rhs = url.split('://', maxsplit=1)
    except ValueError:
        return url if fallback_as_is else None

    rhs_splitted = rhs.split('?', maxsplit=1)
    remainder = rhs_splitted[0]
    query_string = '' if len(rhs_splitted) == 1 else rhs_splitted[1]

    try:
        from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

        if protocol == 'conference':
            return hub_absolute('plainui:' + remainder, query_string=query_string)

        if protocol == 'assembly':
            return hub_absolute('plainui:assembly', query_string=query_string, assembly_slug=remainder)

        if protocol == 'event':
            return hub_absolute('plainui:event', query_string=query_string, event_slug=remainder)

        if protocol == 'wiki':
            return hub_absolute('plainui:static_page', query_string=query_string, page_slug=remainder)

    except NoReverseMatch:
        # we matched a protocol but the remainder was something bogus
        return None

    # no special case encountered, give URL as-is (or return None, if requested)
    return url if fallback_as_is else None


def download_from_url(url: str) -> tuple[str, bytes]:
    # let requests library fetch the URL
    r = requests.get(url, timeout=30)

    # bail out if response is not 200
    r.raise_for_status()

    # try parsing the filename from the Content-Disposition header
    filename = None
    if hdr := r.headers.get('Content-Disposition'):
        # it's surprising that parsing via e.g. EmailMessage works with Wikipedia
        # but fails in unexpected/unrecognizable ways for frab ...
        hdr_split = hdr.split(';', maxsplit=1)
        if len(hdr_split) > 1 and hdr_split[1].startswith('filename='):
            filename = hdr_split[1][len('filename=') :].rsplit('/', maxsplit=1)[-1]

    # if the above fails, try guessing the filename from the URL's path
    if not filename:
        url_parsed = urlparse(url)
        filename = url_parsed.path.rsplit('/')[-1]

    # read the binary content
    r.raw.decode_content = True
    data = BytesIO(r.content)

    # return the result
    return filename, data


def mail2uuid(mail: str, prefix: str = 'acct:', suffix: str = '') -> uuid.UUID:
    uri = prefix + mail + suffix
    return uuid.uuid5(uuid.NAMESPACE_URL, uri)


class GitCloneError(Exception):
    pass


class GitCheckoutError(Exception):
    pass


class GitRepo:
    working_copy_path: Path

    def __init__(self, repo_url):
        self.working_copy_path = None

        # parse given URL and do sanity check (especially, we don't want file:// here)
        parsed = urlparse(repo_url)
        if parsed.scheme not in ['git', 'git+ssh', 'http', 'https']:
            raise ValueError(f'Invalid protocol {parsed.scheme}')

        # note specific commit hash, if given
        self.repo_hash = parsed.fragment if parsed.fragment else None

        # parse optional parameters, if given as query string
        params = parse_qs(parsed.query, strict_parsing=True) if parsed.query else {}
        self.repo_branch = x[0] if (x := params.get('branch')) else x
        self.repo_path = x[0] if (x := params.get('path')) else x

        # note down URL without custom optional parts
        self.repo_url = parsed._replace(query=None, fragment=None).geturl()
        if '$' in self.repo_url:
            raise ValueError("Env variables not allowed in GitRepo's repo_url.")

    def __enter__(self):
        if self.working_copy_path is not None:
            raise NotImplementedError('Accessing GitRepo twice is not supported!')
        self.working_copy_path = Path(tempfile.mkdtemp())

        # assemble git-clone call
        cmd = ['git', 'clone', '-c', 'core.symlinks=false']
        if not self.repo_hash:
            cmd.extend(
                [
                    '--depth',
                    '1',
                ]
            )
        if self.repo_branch:
            cmd.extend(['--branch', self.repo_branch])
        cmd.extend([self.repo_url, str(self.working_copy_path)])

        # this fails with an exception on non-zero exitcode
        logger.info(' '.join(cmd))
        try:
            output = subprocess.run(cmd, shell=False, check=True, capture_output=True)
        except subprocess.CalledProcessError as err:
            raise GitCloneError from err

        logger.debug('Git clone output: %s', output.stdout)

        # do a checkout on the given ref/hash, if given
        if self.repo_hash:
            cmd = ['git', '-C', str(self.working_copy_path), 'checkout', self.repo_hash]
            logger.info(' '.join(cmd))
            try:
                output = subprocess.run(cmd, shell=False, check=True, capture_output=True)
            except subprocess.CalledProcessError as err:
                raise GitCheckoutError from err
            logger.debug('Git checkout output: %s', output.stdout)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not shutil.rmtree.avoids_symlink_attacks:
            raise NotImplementedError('Your OS does not support avoiding symlink attacks.')
        try:
            shutil.rmtree(self.working_copy_path)
        except Exception:
            logger.exception('Failed to remove temporary files.')

    @cached_property
    def base_path(self) -> Path:
        """
        The path on the filesystem on which to base all file operations.
        That path will be using the configured repo_path relative to working copy.
        """
        base_path = Path(self.working_copy_path)
        if self.repo_path:
            base_path = (base_path / self.repo_path).absolute()
            # sanity check that we are not being tricked into some path traversal
            if not base_path.is_relative_to(self.working_copy_path):
                raise ValueError(f'Provided path would result in a directory traversal: {self.repo_path}')

        return base_path

    def get_documents(self, glob: str = '*.md', encoding: str = 'utf-8') -> dict[str, str]:
        """
        Read all documents matching the configured glob as text.
        :param glob: a filter to use when searching the documents to read, defaults to Markdown file extension
        :param encoding: the encoding to use when reading each file, default to UTF-8
        :return: a dictionary mapping filenames to that file's text content
        """
        documents = {}
        for file in self.base_path.rglob(glob):
            # we are only interested in actual files
            if not file.is_file():
                continue

            # read and store files matching the glob
            try:
                documents[file.name] = file.read_text(encoding=encoding)
            except (OSError, UnicodeDecodeError):
                documents[file.name] = None
                logger.exception('Failed to read document "%s".', file.name)

        return documents

    def get_file_content(self, filename: str) -> bytes:
        """
        Read the given file's content from the repository.
        :param filename: relative path (from the configured repo_path within the working copy)
        :return: file content (raw)
        """

        # construct the file's actual path
        file_path = self.base_path / filename

        # sanity check that nobody is trying to do any shenanigans
        if not file_path.is_relative_to(self.working_copy_path):
            raise ValueError(f'Provided file path would result in a directory traversal: {filename}')

        # read the file (or raise if the file does not exist)
        return file_path.read_bytes()
