from jinja2.utils import markupsafe

from django import forms
from django.contrib.auth import forms as auth_forms
from django.forms import ValidationError
from django.utils.formats import localize
from django.utils.http import urlencode
from django.utils.timezone import localtime, now
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from core.abuse import REPORT_CATEGORIES, report_content
from core.base_forms import TranslatedFieldsForm
from core.models import (
    BadgeToken,
    BulletinBoardEntry,
    ConferenceMember,
    ConferenceMemberTicket,
    DirectMessage,
    PlatformUser,
    StaticPage,
    UserBadge,
    UserCommunicationChannel,
)
from core.models.badges import TOKEN_VALIDATOR


class UsernameField(forms.CharField):
    def to_python(self, value):
        user = PlatformUser.objects.filter(username=value).first()
        if user is None:
            raise ValidationError(gettext('Unknown User!'), code='invalid')
        return user


class NewDirectMessageForm(forms.Form):
    def __init__(self, request, conf, user, *args, **kwargs):
        self.request = request
        self.conf = conf
        self.user = user
        super().__init__(*args, **kwargs)

    in_reply_to = forms.UUIDField(required=False)
    recipient = UsernameField(widget=forms.TextInput(attrs={'placeholder': _('Please enter the recipient name')}))
    subject = forms.CharField(max_length=200, min_length=1, strip=True, widget=forms.TextInput(attrs={'placeholder': _('Please enter a subject')}))
    body = forms.CharField(widget=forms.Textarea)

    def clean(self):
        if self.conf.send_pn_disabled:
            raise ValidationError(_('Sending Messages is currently disabled'))
        if self.request.limited:
            raise ValidationError(_('rate-limited'))


class BulletinBoardEntryForm(TranslatedFieldsForm):
    class Meta:
        model = BulletinBoardEntry
        fields = ('title', 'is_public', 'text')

    def __init__(self, request, conference, *args, **kwargs):
        self.request = request
        self.conference = conference
        self.hidden = self.request.user.shadow_banned
        super().__init__(*args, **kwargs)

    title = forms.CharField(max_length=200, min_length=1, strip=True, widget=forms.TextInput(attrs={'placeholder': _('Please enter a title')}))

    def clean(self):
        if not hasattr(self.instance, 'conference'):
            self.instance.conference = self.conference
        if not hasattr(self.instance, 'owner'):
            self.instance.owner = self.request.user
        self.instance.hidden = self.hidden
        if self.conference.board_disabled:
            raise ValidationError(_('Bulletin Board is currently disabled'))
        if self.request.limited:
            raise ValidationError(_('rate-limited'))


class ExampleForm(forms.Form):
    """used in the component gallery"""

    text = forms.CharField(min_length=1, help_text='Help Text', widget=forms.TextInput(attrs={'placeholder': _('Placeholder text')}))
    password = forms.CharField(help_text='Help Text', widget=forms.PasswordInput)
    checkbox = forms.BooleanField(help_text='Help Text')
    textarea = forms.CharField(widget=forms.Textarea, help_text='Help Text')

    def clean(self):
        super().clean()
        if 'text' not in self.cleaned_data:
            raise ValidationError('Some general error message of the form')


class ProfileEditForm(TranslatedFieldsForm):
    class Meta:
        model = PlatformUser
        fields = [
            'pronouns',
            'timezone',
        ]


class ProfileDescriptionEditForm(TranslatedFieldsForm):
    class Meta:
        model = ConferenceMember
        fields = ['description']


class BadgeDefaultsEditForm(TranslatedFieldsForm):
    class Meta:
        model = ConferenceMember
        fields = ['default_badge_visibility']


class RedeemBadgeForm(forms.Form):
    token = forms.CharField(
        required=True,
        label='',
        widget=forms.TextInput(attrs={'placeholder': 'Token'}),
        validators=[TOKEN_VALIDATOR],
        help_text=_('badge__redeem_tokens__token_help'),
    )
    purpose = forms.CharField(required=True, label='', widget=forms.HiddenInput(), initial='redeem_token')
    conference = None

    def __init__(self, *args, **kwargs) -> None:
        self.conference = kwargs.pop('conference', None)
        super().__init__(*args, **kwargs)

    def clean_token(self):
        token = self.cleaned_data['token']
        try:
            badge_token = BadgeToken.objects.get(token=token, badge__conference=self.conference)
        except BadgeToken.DoesNotExist:
            raise ValidationError(_('BadgeToken__does_not_exist %(token)s') % {'token': token})

        return badge_token


class ManageBadgeForm(forms.ModelForm):
    class Meta:
        model = UserBadge
        fields = ['visibility']


class InputTokenForm(forms.Form):
    jwt = forms.CharField(max_length=None, required=True, label=_('Token'))

    def __init__(self, conf, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conf = conf
        self.user = user

    def clean_jwt(self):
        if self.conf.end <= now():
            raise ValidationError(gettext('The Conference is over!'))
        try:
            ConferenceMemberTicket.validate_pretix_ticket(self.conf, self.user, self.cleaned_data['jwt'])
            return self.cleaned_data['jwt']
        except Exception as e:
            raise ValidationError(str(e))


class RedeemTokenAddToUserForm(auth_forms.AuthenticationForm):
    token = forms.CharField(max_length=None, required=True, widget=forms.HiddenInput())

    def __init__(self, conf, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conf = conf

    def clean(self):
        if self.conf.end <= now():
            raise ValidationError(gettext('The Conference is over!'))
        if self.request.limited:
            raise ValidationError(_('rate-limited'))
        return super().clean()


class RedeemTokenUserCreateForm(auth_forms.UserCreationForm):
    class Meta:
        model = PlatformUser
        fields = ('username',)

    token = forms.CharField(max_length=None, widget=forms.HiddenInput())

    def __init__(self, conf, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conf = conf

    def clean(self):
        if self.conf.end <= now():
            raise ValidationError(gettext('The Conference is over!'))
        return super().clean()


class TokenPasswortResetForm(auth_forms.SetPasswordForm):
    jwt = forms.CharField(max_length=None, required=True, widget=forms.HiddenInput())
    username = forms.CharField(required=True, label=_('Username'))

    def __init__(self, conf, user, *args, **kwargs):
        super().__init__(user, *args, **kwargs)
        self.conf = conf
        self.user = user

    def clean(self):
        try:
            ConferenceMemberTicket.validate_pretix_ticket(self.conf, self.user, self.cleaned_data['jwt'], validate_only=True)
        except Exception as e:
            raise ValidationError(str(e))

        if not self.cleaned_data.get('username') or not self.cleaned_data.get('jwt'):
            raise ValidationError(gettext('Please enter the Username that used this Ticket and your new Password'))

        user = PlatformUser.objects.filter(username=self.cleaned_data['username']).first()
        if user is None:
            raise ValidationError(gettext('Unknown User'))

        if not user.is_active:
            raise ValidationError(gettext('User is not active!'))

        if user.email or user.communication_channels.filter(channel=UserCommunicationChannel.Channel.MAIL, is_verified=True).exists():
            raise ValidationError(gettext('User can use password reset by email!'))

        pretix_ident, _ticket_data = ConferenceMemberTicket.validate_pretix_ticket(self.conf, self.user, self.cleaned_data['jwt'], validate_only=True)
        if not ConferenceMemberTicket.objects.filter(conference=self.conf, user=user, ident=pretix_ident).exists():
            raise ValidationError(gettext("User didn't use this Ticket!"))

        self.user = user
        super().clean()


class RoomChoiceField(forms.ModelChoiceField):
    blocked_rooms = {}

    def label_from_instance(self, obj):
        if obj.pk in self.blocked_rooms:
            return markupsafe.Markup('%s [%s]') % (obj, gettext('blocked'))

        return str(obj)


class ReportForm(forms.Form):
    kind = forms.ChoiceField(
        choices=[
            ('url', 'url'),
            ('pn', 'pn'),
            ('board', 'board'),
        ],
        widget=forms.HiddenInput(),
    )
    kind_data = forms.CharField(min_length=1, widget=forms.HiddenInput())
    lookup_key = forms.CharField(widget=forms.HiddenInput(), required=False)
    next = forms.CharField(widget=forms.HiddenInput(), required=False)
    category = forms.ChoiceField(choices=[(k, v[0]) for k, v in REPORT_CATEGORIES.items() if v[2]], initial='content')
    message = forms.CharField(min_length=1, widget=forms.Textarea(), label=_('describe the problem'), required=True)
    message2 = forms.CharField(min_length=1, widget=forms.Textarea(), label=_('describe a solution'), required=True)

    def __init__(self, *args, request, conf, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.conf = conf

    def clean(self):
        if self.request.limited:
            raise ValidationError(_('rate-limited'))
        return super().clean()

    def send_report_mail(self, request):
        from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

        kind = self.cleaned_data['kind']
        moderation_url = None

        if kind == 'url':
            reported_content = self.cleaned_data['kind_data']
            if self.cleaned_data['lookup_key']:
                moderation_url = hub_absolute('backoffice:moderation', query_string=urlencode({'q': self.cleaned_data['lookup_key']}))

        elif kind == 'pn':
            dm = DirectMessage.objects.get(pk=self.cleaned_data['kind_data'])
            reported_content = 'DM by %(sender)s (%(sender_id)s) to %(recipient)s (%(recipient_id)s) at %(timestamp)s (%(uuid)s).\n'
            reported_content += 'Subject: "%(subject)s"\nMessage: "%(body)s"'
            reported_content %= {
                'sender': dm.sender.username,
                'sender_id': dm.sender.id,
                'recipient': dm.recipient.username,
                'recipient_id': dm.recipient.id,
                'timestamp': localize(localtime(dm.timestamp)),
                'uuid': dm.id,
                'subject': dm.subject,
                'body': dm.body,
            }

        elif kind == 'board':
            bbe = BulletinBoardEntry.objects.get(pk=self.cleaned_data['kind_data'])
            reported_content = 'Board Entry by %(owner)s (%(owner_id)s) at %(timestamp)s (%(uuid)s)\n'
            reported_content += 'Title: "%(title)s"\nText: "%(text)s"'
            reported_content %= {
                'owner': bbe.owner.username,
                'owner_id': bbe.owner.id,
                'timestamp': localize(localtime(bbe.timestamp)),
                'uuid': bbe.pk,
                'title': bbe.title,
                'text': bbe.text,
            }
            moderation_url = hub_absolute('backoffice:moderation-board-detail', pk=bbe.pk)

        else:
            raise Exception('Unknown kind')

        report_content(
            request=request,
            reporter=request.user,
            category=self.cleaned_data['category'],
            reported_content=reported_content,
            problem_message=self.cleaned_data['message'],
            proposed_solution=self.cleaned_data['message2'],
            moderation_url=moderation_url,
        )


class StaticPageBodyForm(forms.ModelForm):
    class Meta:
        model = StaticPage
        fields = ['title', 'body']

    title = forms.CharField(min_length=1, required=True, label=_('Wiki__edit__title'), help_text=_('Wiki__edit__title_help'))
    body = forms.CharField(required=True, widget=forms.TextInput(attrs={'rows': 20}), label=_('Wiki__edit__body'), help_text=_('Wiki__edit__body_help'))
    lock = forms.CharField(widget=forms.HiddenInput(), required=False)
