__all__ = (
    'AssembliesEventsView',
    'EventView',
    'SosList',
    'UpcomingView',
)

from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.generic.base import TemplateView

from core.models import (
    Event,
    EventAttachment,
    TagItem,
)
from core.views.list_views import FilteredListView

from plainui.views.utils import ConferenceRequiredMixin, event_filter, session_get_favorite_events


class AssembliesEventsView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/assemblies_events.html.j2'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf

        context['events_upcoming'] = event_filter(self.request.user, self.conf, upcoming=True)
        # not used atm
        context['events_from_assemblies'] = event_filter(self.request.user, self.conf, kinds=[Event.Kind.ASSEMBLY], public_fahrplan=False, calendar_mode=False)
        context['is_favorite_events'] = session_get_favorite_events(self.request.session, self.request.user)

        context['report_info'] = {'enabled': True}

        return context


class EventView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/event.html.j2'

    def get_context_data(self, event_slug, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf

        event = get_object_or_404(Event.objects.conference_accessible(self.conf).select_related('assembly', 'owner', 'track'), slug=event_slug)
        context['event'] = event
        favorites = session_get_favorite_events(self.request.session, self.request.user)
        context['is_favorite'] = str(event.id) in favorites
        context['is_favorite_events'] = favorites
        context['speakers'] = event.public_speakers
        context['tags'] = TagItem.objects.select_related('tag').filter(
            tag__is_public=True, target_type=ContentType.objects.get_for_model(Event), target_id=event.pk
        )
        context['assembly'] = event.assembly
        now = self.now
        if event.schedule_start is None or event.schedule_end is None:
            running_state = ''
        elif now > event.schedule_end:
            running_state = 'complete'
        elif event.schedule_start <= now <= event.schedule_end:
            running_state = 'running'
        else:
            running_state = 'upcoming'
        context['running_state'] = running_state

        context['attachments'] = EventAttachment.objects.filter(
            event=event, visibility__in=[EventAttachment.Visibility.PUBLIC, EventAttachment.Visibility.CONFERENCE]
        )

        context['report_info'] = {'lookup_key': event.slug}
        context['report_info'] = {'url': reverse('plainui:event', kwargs={'event_slug': event.slug})}
        context['fav_info'] = {'type': 'event', 'id': event.id, 'is': str(event.id) in favorites}

        return context


class UpcomingView(ConferenceRequiredMixin, TemplateView):
    template_name = 'plainui/upcoming.html.j2'

    def get_context_data(self, **kwargs):
        events = event_filter(self.request.user, self.conf, upcoming=True)

        return super().get_context_data(
            conf=self.conf,
            events=events,
            my_favorite_events=session_get_favorite_events(self.request.session, self.request.user),
            **kwargs,
        )


class SosList(ConferenceRequiredMixin, FilteredListView):
    model = Event
    template_name = 'plainui/sos.html.j2'
    context_object_name = 'events'
    filters = {}
    paginate_by = 'all'

    def get_queryset(self):
        return event_filter(self.request.user, self.conf, kinds=[Event.Kind.SELF_ORGANIZED], calendar_mode=False)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['conf'] = self.conf

        context['events_upcoming'] = event_filter(self.request.user, self.conf, kinds=[Event.Kind.SELF_ORGANIZED], calendar_mode=False, upcoming=True)
        context['events'] = event_filter(self.request.user, self.conf, kinds=[Event.Kind.SELF_ORGANIZED], calendar_mode=False)
        context['is_favorite_events'] = session_get_favorite_events(self.request.session, self.request.user)
        context['manageable_events'] = Event.objects.manageable_by_user(self.conf, user=self.request.user).only('id')

        context['report_info'] = {'enabled': True}

        return context
